#   Simple Backup - file access management
#
#   Copyright (c)2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
"""
:mod:`sbackup.util.fam` -- file access management
===================================================

.. module:: fam
   :synopsis: public entry point to file access management
.. moduleauthor:: Jean-Peer Lorenz <peer.loz@gmx.net>

"""


from sbackup.util import prefs


_PREFS = prefs.Preferences()
FS_BACKEND = _PREFS.get(key = prefs.PREFS_FS_BACKEND)


# factory methods
def get_fam_target_handler_facade_instance():
    _fam = None

    if FS_BACKEND == prefs.FS_BACKEND_GIO:
        from sbackup.fs_backend import _gio_fam
        _fam = _gio_fam.GioTargetHandler()

    elif FS_BACKEND == prefs.FS_BACKEND_FUSE:
        from sbackup.fs_backend import _fuse_fam
        _fam = _fuse_fam.FuseTargetHandler()

    else:
        raise ValueError("Given filesystem backend `%s` is not supported" % FS_BACKEND)

    return _fam


def get_file_operations_facade_instance():
    _op = None

    if FS_BACKEND == prefs.FS_BACKEND_GIO:
        from sbackup.fs_backend import _gio_utils
        _op = _gio_utils.GioOperations()

    elif FS_BACKEND == prefs.FS_BACKEND_FUSE:
        from sbackup.fs_backend import _fuse_utils
        _op = _fuse_utils.FuseOperations()

    else:
        raise ValueError("Given filesystem backend `%s` is not supported" % FS_BACKEND)

    return _op


def get_remote_services_avail():
    _res = None

    if FS_BACKEND == prefs.FS_BACKEND_GIO:
        from sbackup.fs_backend import _gio_utils
        _res = _gio_utils.REMOTE_SERVICES_AVAIL

    elif FS_BACKEND == prefs.FS_BACKEND_FUSE:
        from sbackup.fs_backend import _fuse_utils
        _res = _fuse_utils.REMOTE_SERVICES_AVAIL

    else:
        raise ValueError("Given filesystem backend `%s` is not supported" % FS_BACKEND)

    return _res

def get_default_remote_service():
    _res = None

    if FS_BACKEND == prefs.FS_BACKEND_GIO:
        from sbackup.fs_backend import _gio_utils
        _res = _gio_utils.REMOTE_SERVICE_SFTP

    elif FS_BACKEND == prefs.FS_BACKEND_FUSE:
        from sbackup.fs_backend import _fuse_utils
        _res = _fuse_utils.REMOTE_SERVICE_SFTP

    else:
        raise ValueError("Given filesystem backend `%s` is not supported" % FS_BACKEND)

    return _res

def get_scheme_from_service(service):
    _res = None

    if FS_BACKEND == prefs.FS_BACKEND_GIO:
        from sbackup.fs_backend import _gio_utils
        _res = _gio_utils.get_scheme_from_service(service)

    elif FS_BACKEND == prefs.FS_BACKEND_FUSE:
        from sbackup.fs_backend import _fuse_utils
        _res = _fuse_utils.get_scheme_from_service(service)

    else:
        raise ValueError("Given filesystem backend `%s` is not supported" % FS_BACKEND)

    return _res


def get_service_from_scheme(scheme):
    _res = None

    if FS_BACKEND == prefs.FS_BACKEND_GIO:
        from sbackup.fs_backend import _gio_utils
        _res = _gio_utils.get_service_from_scheme(scheme)

    elif FS_BACKEND == prefs.FS_BACKEND_FUSE:
        from sbackup.fs_backend import _fuse_utils
        _res = _fuse_utils.get_service_from_scheme(scheme)

    else:
        raise ValueError("Given filesystem backend `%s` is not supported" % FS_BACKEND)

    return _res



#   Simple Backup - Operation System related utilities
#
#   Copyright (c)2010,2013: Jean-Peer Lorenz <peer.loz@gmx.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

# UNIX specific implementations/definitions

import os
import pwd
import grp
import stat
import subprocess
import types
import atexit
import signal

import glib


PATHSEP = os.sep
UNIX_PERM_ALL_WRITE = stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH
UNIX_PERM_GRPOTH_NORWX = ~(stat.S_IRWXG | stat.S_IRWXO)


COMMAND_GREP = "grep"
COMMAND_PS = "ps"


# default values for environment variable if not set
# see: http://standards.freedesktop.org/basedir-spec/basedir-spec-0.6.html
# Note: do not remove trailing slashs, it does not work without them!
ENVVAR_XDG_DATA_DIRS = "XDG_DATA_DIRS"
DEFAULT_XDG_DATA_DIRS = "/usr/local/share/:/usr/share/"

ENVVAR_PATH = "PATH"
DEFAULT_PATH = "/usr/bin:/usr/local/bin:/bin"


CLEAN_ENVIRONMENT = {
# default Gnome environment 
    "_" : "",
    "COLORTERM" : "",
    "DBUS_SESSION_BUS_ADDRESS" : "",
    "DEFAULTS_PATH" : "",
    "DESKTOP_SESSION" : "",
    "DISPLAY" : "",
    "GDM_KEYBOARD_LAYOUT" : "",
    "GDMSESSION" : "",
    "GDM_LANG" : "",
    "GNOME_KEYRING_PID" : "",
    "GNOME_KEYRING_CONTROL" : "",
    "GPG_AGENT_INFO" : "",
    "GTK_MODULES" : "",
    "HISTCONTROL" : "",
    "HOME" : "",
    "LANG" : "",
    "LC_ALL" : "",
    "LD_LIBRARY_PATH" : "",
    "LESSCLOSE" : "",
    "LESSOPEN" : "",
    "LOGNAME" : "",
    "MANDATORY_PATH" : "",
    "MANPATH" : "",
    "OLDPWD" : "",
    "PATH" : "",
    "PWD" : "",
    "PYTHONPATH" : "",
    "SESSION_MANAGER" : "",
    "SHELL" : "",
    "SHLVL" : "",
    "SPEECHD_PORT" : "",
    "SSH_AUTH_SOCK" : "",
    "TERM" : "",
    "USER" : "",
    "USERNAME" : "",
    "WINDOWID" : "",
    "XAUTHORITY" : "",
    "XDG_CONFIG_DIRS" : "",
    "XDG_DATA_DIRS" : "",
    "XDG_SESSION_COOKIE" : "",
# additional KDE environment variables
    "KDE_FULL_SESSION" : "",
    "GS_LIB" : "",
    "DM_CONTROL" : "",
    "SSH_AGENT_PID" : "",
    "XDM_MANAGED" : "",
    "KDE_SESSION_VERSION" : "",
    "LD_BIND_NOW" : "",
    "GTK2_RC_FILES" : "",
    "WINDOWPATH" : "",
    "XCURSOR_THEME" : "",
    "KDE_SESSION_UID" : "",
    "QT_PLUGIN_PATH" : ""
    }


def get_user_home_dir():
    return os.path.expanduser('~')


def get_user_config_dir():
    _confdir = glib.get_user_config_dir()
    return _confdir


def get_user_data_dir():
    _datadir = glib.get_user_data_dir()
    return _datadir


def get_pid():
    _pid = os.getpid()
    return _pid


def get_user_from_env():
    """Returns the USER defined in current environment. If no USER is
    set, None is returned.
    """
    _user = os.environ.get("USER", None)
    return _user


def get_user_from_uid():
    _user = pwd.getpwuid(os.getuid()).pw_name
    return _user


def is_superuser():
    _res = False
    if os.geteuid() == 0:
        _res = True
    return _res


def switch_user(uid_name):
    """
    """
    starting_uid = os.getuid()
#    starting_gid = os.getgid()
#    starting_uid_name = pwd.getpwuid(starting_uid)[0]
#    print 'switch_user: started as %s/%s' % \
#    (pwd.getpwuid(starting_uid).pw_name,
#    grp.getgrgid(starting_gid).gr_name)
#    print "UID: %s  EUID: %s  GID: %s" % (os.getuid(), os.geteuid(), os.getgid())

    if os.getuid() != 0:
        # We're not root so, like, whatever dude
#        print "switch_user: already running as '%s'" % starting_uid_name
        return

    # If we started as root, drop privs and become the specified user/group
    if starting_uid == 0:
        # Get the uid/gid from the name
        running_uid = pwd.getpwnam(uid_name).pw_uid
        running_gid = pwd.getpwnam(uid_name).pw_gid

    # Try setting the new uid/gid
    try:
        os.setgid(running_gid)
    except OSError, e:
        print 'Could not set effective group id: %s' % e

    try:
        os.setgroups([running_gid])
    except OSError, e:
        print 'Could not set associated groups: %s' % e

    try:
        os.setuid(running_uid)
    except OSError, e:
        print 'Could not set effective user id: %s' % e

    # Ensure a very convervative umask
    new_umask = 077
    old_umask = os.umask(new_umask)
#    print 'drop_privileges: Old umask: %s, new umask: %s' % \
    (oct(old_umask), oct(new_umask))

    final_uid = os.getuid()
    final_gid = os.getgid()
    print 'drop privileges: running as %s/%s' % \
    (pwd.getpwuid(final_uid).pw_name,
    grp.getgrgid(final_gid).gr_name)


def drop_privileges():
    switch_user(uid_name = 'nobody')


def set_grp(groupname):
    # [Bug 112540] : Let the admin group have read access to the backup dirs
    if os.geteuid() == 0:
        try:
            _gid = grp.getgrnam(groupname).gr_gid
            os.setgid(_gid)
        except KeyError:
            # LP #696183:
            print "Group not changed to `%s`: unknown group" % groupname
        except OSError, error:
            print "Failed to set GID to `%s`: %s" % (groupname, error)


def nice():
    os.nice(5)


def very_nice():
    os.nice(20)


def get_environment():
    _env = os.environ
    return _env


def set_default_environment():
    """Sets required environment variables to their specified default
    values if not defined. This can happen e.g. some root shells where
    no environment variable for the freedesktop.org base directories
    are defined.
    
    """
    _vars = { ENVVAR_XDG_DATA_DIRS : DEFAULT_XDG_DATA_DIRS,
             ENVVAR_PATH : DEFAULT_PATH
           }
    for var in _vars:
        val = os.environ.get(var)
        if val is None:
            os.environ[var] = _vars[var]


def get_process_environment(pid):
    _envlst = None
    _envfile = "/proc/%s/environ" % pid
    try:
        fobj = open(_envfile, "rb")
        _envlst = [_var for _var in _readline_nullsep(fobj)]
        fobj.close()
    except (OSError, IOError), error:
        print "Unable to open Gnome session environment: %s." % error
        _envlst = None

    _env = None
    if _envlst is not None:
        _env = {}
        for _var in _envlst:
            if _var != "":
                _vars = _var.split("=", 1)
                if len(_vars) != 2:
                    print "Unable to read environment variable '%s'. Skipped." % _var
                else:
                    _env[_vars[0]] = _vars[1]
    return _env


def get_clean_environment():
    _clean_env = {}
    _env = get_environment()
    for _var in CLEAN_ENVIRONMENT:
        _val = _env.get(_var, "")
        if _val != "":
            _clean_env[_var] = _val
    return _clean_env


def set_gio_env_from_session():
    _set_envvar_from_session(key = 'SSH_AUTH_SOCK')
    _set_envvar_from_session(key = 'GNOME_KEYRING_CONTROL')
    _set_envvar_from_session(key = 'GNOME_KEYRING_PID')
    _set_envvar_from_session(key = 'XAUTHORITY')
    _set_envvar_from_session(key = 'GPG_AGENT_INFO')
    #===========================================================================
    # key_kr = 'GNOME_KEYRING_CONTROL'
    # key_ssh = 'SSH_AUTH_SOCK'
    # _value = os.environ.get(key_kr)
    # if _value is None:
    #    _session = get_session_name()
    #    _env = get_session_environment(_session)
    #    if _env is not None:
    #        _nvalue = _env.get(key_kr)
    #        if _nvalue is not None:
    #            _nvalue = "%s/ssh" % _nvalue
    #            print "Updating %s to: %s" % (key_ssh, _nvalue)
    #            os.environ[key_ssh] = _nvalue
    #===========================================================================


def is_dbus_session_bus_set():
    key = 'DBUS_SESSION_BUS_ADDRESS'
    _res = False
    _value = os.environ.get(key)
    if _value is not None:
        _res = True
    return _res


def launch_dbus_if_required():
    # launches a dbus session bus for GIO
    #
    # inspired by code from Duplicity/giobackend
    # originally written by Michael Terry (2009)
    #
    if not is_dbus_session_bus_set():
        output = exec_command_stdout(['dbus-launch'])
        print "D-Bus session bus launched"
        for _line in output.split('\n'):
            print "\t%s" % _line
            _var = _line.split('=', 1)
            if len(_var) != 2:
                print "Unable to read environment variable '%s'. Skipped." % _line
            else:
                os.environ[_var[0]] = _var[1]
                if _var[0] == 'DBUS_SESSION_BUS_PID':
                    # we need to kill the launched bus at termination
                    try:
                        _pid = int(_var[1])
                        atexit.register(os.kill, _pid, signal.SIGTERM)
                    except ValueError, error:
                        print "Unable to register dbus clean-up action: %s" % error


def set_dbus_session_bus_from_session():
    """Update dbus session bus address in order to connect to gvfsd from root consoles
    """
    _set_envvar_from_session(key = 'DBUS_SESSION_BUS_ADDRESS')


def set_display_from_session():
    _set_envvar_from_session(key = 'DISPLAY')


def _set_envvar_from_session(key):
    _value = os.environ.get(key)
    if _value is None:
        _env = get_session_environment()
        if _env is not None:
            _nvalue = _env.get(key)
            print "Updating %s to: %s" % (key, _nvalue)
            if _nvalue is not None:
                os.environ[key] = _nvalue


#TODO: implement Singleton class for access to environment
def get_session_environment():
    #FIXME: support Unity, Mate, Cinnamon
    _sessionnames = ("x-session-manager", "gnome-session", "gnome-shell",
                     "ksmserver", "xfce4-session")
    mod_env = None
    for _name in _sessionnames:
        if grep_pid(_name) is not None:
            mod_env = _get_session_env(session = _name)
            print "Desktop session %s found" % _name
            break
    if mod_env is None:
        print "No Desktop session found"
    return mod_env


def _get_session_env(session):
    _mod_env = None
    _session_pid = grep_pid(processname = session)
    if _session_pid is None:
        print "Session `%s` not found" % session
    else:
#        print "Session `%s` PID: %s" % (session, _session_pid)
        _mod_env = get_clean_environment()
        _session_env = get_process_environment(pid = _session_pid)
        if _session_env is None:
            _mod_env = None
        else:
            _mod_env.update(_session_env)
#    debug_print_environment(_mod_env, "modified from desktop session")
    return _mod_env


def debug_print_environment(env, description):
    print "\nEnvironment `%s`:" % description
    print "-----------------------------------------------------------------------------"
    for _var in env:
        print "%s: %s" % (_var, env[_var])
    print "-----------------------------------------------------------------------------"


def exec_command_async(args, env = None):
    if not isinstance(args, types.ListType):
        raise TypeError("List of arguments expected.")
    pid = subprocess.Popen(args, env = env).pid
    return pid


def exec_command_stdout(args, env = None):
    """
    :note: Standard output is hold in memory. Don't use this if you expect much output.
    """
    if not isinstance(args, types.ListType):
        raise TypeError("List of arguments expected.")
    _output = subprocess.Popen(args, stdout = subprocess.PIPE, env = env).communicate()[0]
    _output = _output.strip()
    return _output


def exec_command_returncode(args, env = None):
    """
    """
    if not isinstance(args, types.ListType):
        raise TypeError("List of arguments expected.")
    _ret = subprocess.call(args, env = env)
    return _ret


def grep_pid(processname):
    """
    note for using pgrep: The process name used for matching is limited to
    the 15 characters present in the output of /proc/pid/stat. Use
    the -f option to match against the complete command line /proc/pid/cmdline.
    
    therefore here a full RegEx is applied to complete command lines:
    pgreg -f "^(/(.+/)*)?processname($|\s+.+)
    
    """
    _pid = None
    output = exec_command_stdout(args = ["pgrep", "-f",
                                    "^(/(.+/)*)?%s($|\s+.+)" % processname])
    print "DEBUG: pgrep %s = %s" % (processname, output)
    try:
        _pid = int(output)
    except ValueError:
#        print "Unable to get PID of process '%s'." % processname
        _pid = None
    return _pid


def pid_exists(pid, processname = None):
    """
    @type pid: String
    """
    if not isinstance(pid, types.StringTypes):
        raise TypeError("PID expected as string.")
    if processname is not None:
        if not isinstance(processname, types.StringTypes):
            raise TypeError("Process name expected as string.")

    _exists = False
    output = exec_command_stdout(args = [COMMAND_PS, "--no-headers", "-lF", "--pid", pid])
    if output != "":
        if processname is None:
            _exists = True
        else:
            if processname in output:
                _exists = True
    return _exists


def proc_exists(processname, env = None):
    if not isinstance(processname, types.StringTypes):
        raise TypeError("Process name expected as string.")
    if COMMAND_GREP in processname:
        raise ValueError("Name of checked process must not contain `%s`" % COMMAND_GREP)

    _exists = False
    cmd_ps = subprocess.Popen([COMMAND_PS, "ax", "--no-headers", "-lF"],
                              stdout = subprocess.PIPE, env = env)
    cmd_v_grep = subprocess.Popen([COMMAND_GREP, "-v", COMMAND_GREP],
                                  stdin = cmd_ps.stdout, stdout = subprocess.PIPE, env = env)
    cmd_grep = subprocess.Popen([COMMAND_GREP, processname],
                                stdin = cmd_v_grep.stdout, stdout = subprocess.PIPE, env = env)
    _output = cmd_grep.communicate()[0]
    _output = _output.strip()
    if _output != "":
        _exists = True
    return _exists


def _readline_nullsep(fd):
    """
    Iterator that read a NUL separeted file as lines 
    @param fd: File descriptor
    @return: the gotten line
    @rtype: String
    """
    _continue = True

    while _continue is True:
        c = fd.read(1)
        currentline = ''

        while c:
            if c == '\0'  :
                # we got a line
                break
            currentline += c
            c = fd.read(1)
        else:
            # c is None
            _continue = False

        yield currentline




#   Simple Backup - configuration file handling
#
#   Copyright (c)2009-2010,2013: Jean-Peer Lorenz <peer.loz@gmx.net>
#   Copyright (c)2007-2009: Ouattara Oumar Aziz <wattazoum@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

"""
Configuration definition

For a description of sbackup's file format please have a look at the shipped
example configuration file in sbackup/data/sbackup.conf.example.

"""


from gettext import gettext as _
import ConfigParser
import smtplib
import types
import os.path
import re

from sbackup.pkginfo import Infos

from sbackup import util

from sbackup.util import local_file_utils
from sbackup.util import exceptions
from sbackup.util import pathparse
from sbackup.util import log
from sbackup.util import system


_CRON_PATH_TEMPLATE = "/etc/cron.%s/%s"
_LAUNCHER_NAME_CRON = "sbackup"
_ANACRON_SERVICES = ["hourly", "daily", "weekly", "monthly"]
_CRON_SERVICE = "d"

_SCHEDULE_SERVICES = _ANACRON_SERVICES
_SCHEDULE_SERVICES.append(_CRON_SERVICE)

SCHEDULE_TYPE_ANACRON = 0
SCHEDULE_TYPE_CRON = 1
SCHEDULE_TYPES = [SCHEDULE_TYPE_CRON, SCHEDULE_TYPE_ANACRON]


def is_default_profile(conffile):
    """Checks whether the given configuration file corresponds to the
    default profile.
    
    @param conffile: full path to configuration file to check
    @return: True if the file corresponds to the default profile, otherwise False.
    @rtype: Boolean
    
    """
    is_default = False
    cfile = os.path.basename(conffile)

    if cfile == ConfigManagerStaticData.get_default_conffile():
        is_default = True

    return is_default


def get_logfile_name_template(conffile):
    """Determines the profilename from the given pathname `conffile`
    and returns an appropriate logfile name.
    
    By 'template' is indicated this is the base logfile name. It is
    extended by the current time if a logfile is actually created. 
    
    The given `conffile` may not exist.
         
    """
    conffile_hdl = ConfigurationFileHandler()
    profilename = conffile_hdl.get_profilename(conffile)
    if profilename == ConfigManagerStaticData.get_default_profilename():
        logfname = ConfigManagerStaticData.get_default_logfile()
    else:
        logfname = ConfigManagerStaticData.get_profile_logfile(profilename)
    return logfname

def get_profiles(prfdir):
    """Get the configuration profiles list
     
    @return: a dictionarity of { name: [path_to_conffile, enable] } 
    """
    return _get_profiles(prfdir, application = "sbackup")

def get_profiles_nssbackup(prfdir):
    """Get the configuration profiles list
     
    @return: a dictionarity of { name: [path_to_conffile, enable] } 
    """
    return _get_profiles(prfdir, application = "nssbackup")

def _get_profiles(prfdir, application):
    """Get the configuration profiles list
     
    @return: a dictionarity of { name: [path_to_conffile, enable] } 
    """
    profiles = dict()
    if os.path.exists(prfdir) and os.path.isdir(prfdir):
        _listing = os.listdir(prfdir)
        if application == "nssbackup":
            _re = ConfigManagerStaticData.get_profilename_nssbackup_re()
        else:
            _re = ConfigManagerStaticData.get_profilename_re()
        for conff in  _listing:
            mobj = _re.match(conff)
            if mobj is not None:
                name = mobj.group(1)
                path = os.path.join(prfdir, mobj.group(0))
                if os.path.isfile(path):
                    enable = (mobj.group(2) is None)
                    profiles[name] = [path, enable]
    return profiles


def get_default_config_obj():
    """Creates and returns an object containing the default configuration.
    Differences between normal users and privileged users (i.e. admins) are taken into account.
    """
    if system.is_superuser():
        _default_config = _DefaultConfigurationForAdmins()
    else:
        _default_config = _DefaultConfigurationForUsers()
    return _default_config


def parse_cronexpression(cronfile_content):
    _expr = None
    for _line in cronfile_content.split("\n"):
        _line = _line.strip()
        _line = " ".join(_line.split()) # eliminate multiple spaces and tabs
        if _line.startswith("#"):
            continue

        elif _line.startswith("@hourly"):
            _expr = "0 * * * *"
            break
        elif _line.startswith("@daily"):
            _expr = "0 0 * * *"
            break
        elif _line.startswith("@weekly"):
            _expr = "0 0 * * 0"
            break
        elif _line.startswith("@monthly"):
            _expr = "0 0 1 * *"
            break
        elif _line.startswith("@"):
            _expr = None    # not supported (@yearly,...)
            break

        else:
            _tup = _line.split(" ", 6)  # now split at spaces
            if len(_tup) == 7:
                _expr = " ".join(_tup[0:5])
                break
    return _expr





class ConfigManager(ConfigParser.ConfigParser):
    """SBackup config manager
    
    The configuration manager is responsible for the following:
     
    * creates a logger instance with specified log level and log file target
    
    @todo: The configuration manager should not create the logger itself!
           This should be done outside of the configuration after reading and
           parsing the config file.
           
    @todo: When a configuration is loaded from disk/file this configuration
            is checked for every required option. Non-existing options are
            then filled with reasonable default values.
            Default values should never be set from the GUI nor should
            default values are taken as 'implicitly' given in any method
            that requires such setting.
            This is not implemented yet.
            
    @todo: Use the RawConfigParser instead of ConfigParser! And maybe the \
            parser should be a member of SBackup config!
    """

    def __init__(self, configfile = None):
        """Default constructor.
        
        @param configfile: Full path to the used configuration file.
        
        @todo: remove command-line parsing from here; \
                this class should only take a filename as parameter.
        """
        ConfigParser.ConfigParser.__init__(self)
        self.__conffile_hdl = ConfigurationFileHandler()

        self.__servicefile = ConfigManagerStaticData.get_schedule_script_file()
        self.__schedules_user = "root"    # schedules are only usable by root 

        # configuration object which is set to valid default values
        self.__default_config = _DefaultConfiguration() # is a dummy
        self.__create_default_config_obj()

        self.__dirconfig = {}

        self.__logfile_dir = ""
        self.__current_logfile = None
        self.conffile = None
        self.logger = None
        self.__profileName = None

        self.valid_options = {}
        self.filename_from_argv = None
        self.argv_options = {}

        self.setValidOpts(ConfigManagerStaticData.get_our_options())

        # use the given conf-file only if no was given on cmdline
        if not self.conffile and configfile:
            self.conffile = configfile

        _conffile_used = False    # helper to print out an informative message

        # if a conf-file is set, evaluate this one and overwrite default values
        if self.conffile:
            self.read(self.conffile)
            _conffile_used = True
        else:
            self._set_defaultprofile_values_to_default()

        self.__create_logger()

        self.validateConfigFileOpts()
        self.__check_destination_option()

        if _conffile_used:
            self.logger.info(_("Profile settings are being read from file '%s'.")\
                              % self.conffile)
        else :
            self.logger.info(_("Profile settings are being set to default values. Configuration file is set to '%s'.")\
                            % self.conffile)

    def __del__(self):
        self.__cleanup_logfiles()

    def __str__(self):
        retval = []
        for section, sec_data in self._sections.iteritems():
            retval.append("[%s]" % section)
            [retval.append("%s = %s" % (o, repr(v)))
                for o, v in sec_data.items() if o != '__name__']
        return "\n".join(retval)

    def __init_sections(self):
        """Initializes the sections of the configuration.
        
        @todo: Use the 'our_options' from ConfigManagerStaticData class for \
                initialisation.
                
        @note: Should we remove *unknown* sections? No, though we validate the file.
        """
        if not self.has_section("general"):
            self.add_section("general")
        if not self.has_section("dirconfig"):
            self.add_section("dirconfig")
        if not self.has_section("exclude"):
            self.add_section("exclude")
        if not self.has_section("log"):
            self.add_section("log")
        if not self.has_section("report"):
            self.add_section("report")
        if not self.has_section("places"):
            self.add_section("places")
        if not self.has_section("schedule"):
            self.add_section("schedule")

    def _set_defaultprofile_values_to_default(self):
        """Makes this configuration the default profile by pointing it to
        the default configuration file and sets according default values
        for the configuration. It distinguishes between users and super-users.
        """
        self.conffile = self.__conffile_hdl.get_default_conffile_fullpath()
        self.set_values_to_default()

    def __create_default_config_obj(self):
        """Creates an object containing the default configuration and stores it
        in the according instance variable. Differences between normal users
        and privileged users (i.e. admins) is taken into account.
        """
        self.__default_config = get_default_config_obj()

    def __get_default_config_obj(self):
        """Private accessor method that returns the default configuration
        object. Use this method rather direct access to the instance
        variable in order to ensure the object exists.
        """
        if self.__default_config is None:
            raise ValueError("No default configuration object was created yet.")
        if not isinstance(self.__default_config, _DefaultConfiguration):
            raise TypeError("Given configuration is not of type \
                            'DefaultConfiguration'. Got '%s' instead." \
                            % type(self.__default_config))
        return self.__default_config

    def set_values_to_default(self):
        """Sets default values for this configuration. It is distinguished
        between normal users and super-users. The path to the current
        configuration file (i.e. the current profile) is not touched.
        
        @note: This method's purpose is to provide reasonable values when
                creating a fresh configuration and restoring of these
                'recommended' (or predefined) defaults on existing configs.
        """
        defaults = self.__get_default_config_obj()
        self.__init_sections()
        # General
        self.set("general", "maxincrement", str(defaults.get_max_increment()))
        self.set("general", "format", defaults.get_compress_format())
        self.set("general", "splitsize", str(defaults.get_split_size()))

        # dirconfig and excludes
        self.__set_dirconfig(defaults.get_dir_config())
        self.__set_regex_excludes(defaults.get_regex_excludes())

        # other exclude reasons
        self.set("exclude", "maxsize", str(defaults.get_max_filesize()))
        followlinks = "0"
        if defaults.get_follow_links() is True:
            followlinks = "1"
        self.set("general", "followlinks", followlinks)

        # target (= destination)
        self.set_target_to_default()

        # schedule
        schedule = defaults.get_schedule()
        if schedule[1] != "":
            is_cron = 0
            if schedule[0] is True:
                is_cron = 1
            self.setSchedule(is_cron, schedule[1])

        # Purging
        self.set("general", "purge", defaults.get_purge())

        # Section log
        self.set("log", "level", str(defaults.get_loglevel()))
        self.set_logdir(defaults.get_logdir())
        self.set_logfile_templ_to_config()
        if not local_file_utils.path_exists(self.get("log", "file")) :
            local_file_utils.createfile(self.get("log", "file"))

        # report settings
        self.__clear_report_section()
        # LP Bug #153605
# TODO: we should not set a default 'from' since the value is not usable
# for 99% of the users (despite LP Bug #153605)
        self.set("report", "from", defaults.get_report_smtpfrom())

        # remaining administrative settings
        self.set("general", "mountdir", defaults.get_mountdir())
        self.set("general", "lockfile", defaults.get_lockfile())
        self.set("places", "prefix", defaults.get_prefix())

    def set_target_to_default(self):
        """The destination (target) option is set to the default value
        valid for the current user. If the specified path does not
        exist is a directory created.
        """
        _def_target_path = self.get_target_default()
        self.set("general", "target", _def_target_path)
        if not os.path.exists(_def_target_path):
            os.mkdir(_def_target_path)

    def get_target_default(self):
        """Wrapper method that return the default target path for
        the current user. No values of this configuration (actually
        configuration manager) are touched. 
        """
        defaults = self.__get_default_config_obj()
        target = defaults.get_target()
        return target

    def __check_destination_option(self):
        if not self.has_option("general", "target"):
            raise exceptions.SBException (_("Option 'target' is missing, aborting."))

    def get_destination_path(self):
        """Returns the target option that is currently set. Purpose is to
        hide the actual naming of options from external classes.
        Returns None if not set.  
        """
        self.__check_destination_option()
        _path = self.get("general", "target")
        _uri = pathparse.UriParser()
        _uri.set_and_parse_uri(uri = _path)     # normalizes and validates given path
        _res = _uri.uri
        return _res

#    def get_destination_obj(self):
#        self.__check_destination_option()
#        _dest = self.get("general", "target")
#        _dest_obj = pathparse.UriParser()
#        _dest_obj.set_and_parse_uri(uri = _dest)
#        return _dest_obj

    def get_mountdir(self):
        _defaults = self.__get_default_config_obj()
        _mdir = _defaults.get_mountdir()

        if self.has_option("general", "mountdir"):
            _mdir = self.get("general", "mountdir")
        return _mdir

    def has_maxsize_limit(self):
        """Returns True if the 'maxsize' option is enabled (i.e. set to
        any value greater than 0).
        """
        _res = False
        if self.get_maxsize_limit() > 0:
            _res = True
        return _res

    def get_maxsize_limit(self):
        """Returns the maximum file size limit. If the option is not set, 0 is returned.
        """
        _section = "exclude"
        _option = "maxsize"
        _maxsize = 0
        if self.has_option(_section, _option):
            _val = int(self.get(_section, _option))
            if _val > 0:
                _maxsize = _val
        return _maxsize

    def get_dirconfig_local(self):
        """Returns a list of pairs of (name, value) from the 'dirconfig' section.
        If no values are set, None is returned.
        """
        _section = "dirconfig"
        _res = None
        if self.has_section(_section):
            _items = self.items(_section)
            if len(_items) > 0:
                _res = []
                for _item in _items:
                    assert len(_item) == 2
                    if _item[0] == "remote":
                        pass
                    else:
                        try:
                            _res.append((_item[0].replace("\\x3d", "="), int(_item[1])))
                        except ValueError, error:
                            _msg = _("Error while processing configuration:\n\n"\
                                     "%(error)s in line `%(line)s`.\n\n"\
                                     "Please check your configuration file.") % {"error" : str(error),
                                                                                 "line" : _item[0]}
                            raise exceptions.NonValidOptionException(_msg)
        return _res

    def get_followlinks(self):
        _section = "general"
        _option = "followlinks"
        _res = False
        if self.has_option(_section, _option):
            if str(self.get(_section, _option)) == "1":
                _res = True
        return _res

    def optionxform(self, option):
        """Default behaviour of ConfigParser is to set the option
        keys to lowercase. By overiding this method, we make it
        case sensitive. that's really important for dirconfig paths. 
        """
        return str(option)

    def has_option(self, section, option):
        """Checks this configuration for a given option comprising of
        section name and option name.
        
        @rtype: Boolean
        """
        if section == "dirconfig" and \
                not ConfigParser.ConfigParser.has_option(self, section, option):
            if option == "remote" :
                return ConfigParser.ConfigParser.has_option(self, section,
                                                            option)
            #search through remote option to get the option
            if ConfigParser.ConfigParser.has_option(self, "dirconfig",
                                                            'remote'):
                remotes = self.get("dirconfig", "remote")
                if type(remotes) == str :
                    remotes = eval(remotes)
                if type(remotes) != dict :
                    raise exceptions.SBException(_("Unable to evaluate '%(parameter)s' as a dictionary (value got = '%(value)r').")\
                            % {'parameter': remotes, 'value': type(remotes)})
                if not remotes.has_key(option) :
                    # then it wasn't for us , fall back on the parent
                    return ConfigParser.ConfigParser.has_option(self, section,
                                                                option)
                else :
                    # we have this key
                    return True
            else :
                return ConfigParser.ConfigParser.has_option(self, section,
                                                            option)
        else :
            #fall back in parent behaviour
            return ConfigParser.ConfigParser.has_option(self, section, option)

    def get(self, section, option):
        """Returns a given option value from this config.
        """
        # if we have (dirconfig,opt), if opt=remote
        if section == "dirconfig" and not option == 'remote' and self.has_option(section, option) and not ConfigParser.ConfigParser.has_option(self, section, option):
            #search through remote option to get the option
            remotes = ConfigParser.ConfigParser.get(self, "dirconfig", "remote", True)
            if type(remotes) == str :
                remotes = eval(remotes)
            if type(remotes) != dict :
                raise exceptions.SBException(_("Couldn't evaluate '%(parameter)s' as a dictionary (value got = '%(value)r' )") % {'parameter': remotes, 'value': type(remotes)})
            # we have that key
            return remotes[option]
        elif section == "dirconfig" and option == 'remote' and ConfigParser.ConfigParser.has_option(self, section, option):
            remotes = ConfigParser.ConfigParser.get(self, "dirconfig", "remote", True)
            if type(remotes) == str :
                remotes = eval(remotes)
            if type(remotes) != dict :
                raise exceptions.SBException(_("Unable to evaluate '%(parameter)s' as a dictionary (value got = '%(value)r').") % {'parameter': remotes, 'value': type(remotes)})
            return remotes
        else :
            #fall back in parent behaviour
            return ConfigParser.ConfigParser.get(self, section, option, True)

    def set(self, section, option, value):
        """Set an option just like a configParser but in case of
        the 'remote' option in 'dirconfig'. In this case, value
        must be a dict with the value you want to set, e.g.
        value = {'ssh://test/': 1, 'ssh://test/test': 0}
        You can set one at a time, the value will be append to
        the 'remote' dict.
        """
        if section == "dirconfig" and option == "remote" :
            if type(value) != dict :
                raise exceptions.SBException(_("You must provide a dictionary."))
            if not self.has_option(section, option) :
                ConfigParser.ConfigParser.set(self, section, option, value)
            else :
                remotes = ConfigParser.ConfigParser.get(self, section, option, True)
                if type(remotes) == str :
                    remotes = eval(remotes)
                if type(remotes) != dict :
                    raise exceptions.SBException("Couldn't eval '%s' as a dict (value got = '%r' )" % (remotes, type(remotes)))
                for rsource, flag in value.iteritems() :
                    remotes[rsource] = flag
                ConfigParser.ConfigParser.set(self, section, option, remotes)
        else :
            #fall back in normal bahaviour
            ConfigParser.ConfigParser.set(self, section, option, value)

    def remove_option(self, section, option):
        """Remove an option, but it's different for remote.
        If option = "remote" then the whole remote option will be removed.
        If option is in remote dict, section ='dirconfig' and
        option='ssh://test/me' then the entry in the remote dict will be removed.
        """
        if section == "dirconfig" and not ConfigParser.ConfigParser.has_option(self, section, option) :
            #search through remote option to get the option
            if not self.has_option("dirconfig", "remote"):
                #fall back in parent behaviour
                ConfigParser.ConfigParser.remove_option(self, section, option)
            else :
                self.logger.debug("search through remote option to get the option")
                remotes = self.get("dirconfig", "remote")
                if type(remotes) == str :
                    remotes = eval(remotes)
                if type(remotes) != dict :
                    raise exceptions.SBException("Couldn't eval '%s' as a dict (value got = '%r' )" % (remotes, type(remotes)))
                if not remotes.has_key(option) :
                    # then it wasn't for us , fall back on the parent
                    ConfigParser.ConfigParser.remove_option(self, section, option)
                else :
                    # we have that key
                    remotes.pop(option)
                    self.logger.debug("remote is now '%r'" % remotes)
                    ConfigParser.ConfigParser.set(self, section, "remote", remotes)
        else :
            #fall back in parent behaviour
            ConfigParser.ConfigParser.remove_option(self, section, option)

    def __clear_dirconfig(self):
        """The internal variable containing the directory configuration
        (i.e. what directories are being included resp. excluded) is
        cleared. The according config section remains but is empty then.
        """
        _section = "dirconfig"
        self.__dirconfig.clear()
        if self.has_section(_section):
            self.remove_section(_section)
        if not self.has_section(_section):
            self.add_section(_section)

    def __set_dirconfig(self, dirconf):
        """The configuration item 'dirconfig' (Directory configuration) is
        set to the value of the given dictionary. Previous values are
        overwritten.
        
        @param dirconf:  new value
    
        @type dirconf:      Dictionary
        
        @return: None
        
        @raise TypeError: If the given parameter is not of dictionary type

        """
        _section = "dirconfig"
        if not isinstance(dirconf, types.DictionaryType):
            raise TypeError("Given parameter must be a Dictionary. "\
                            "Got %s instead." % (type(dirconf)))

        self.__clear_dirconfig()
        self.__dirconfig.update(dirconf)
        for a, b in self.__dirconfig.iteritems() :
            self.set(_section, a, b)

    def __set_regex_excludes(self, aregex):
        """Helper method that sets the excludes defined by Regular
        Expressions.
        """
        _section = "exclude"
        _option = "regex"
        if not isinstance(aregex, types.StringTypes):
            raise TypeError("Given parameter must be of string type. "\
                            "Got %s instead." % (type(aregex)))
        self.set(_section, _option, aregex)

    def setValidOpts(self, valid_options, parse_cmdline = False):
        self.valid_options = valid_options

    def __create_logger(self):
        """Initializes logger with profile name as identifier
        and use the specified file as log file.
        
        Uses the defined 'template' which is extended by the current time
        in order to provide a unique name when the file is actually created. 
        """
        curr_logf = None
        if self.has_section("log") and self.has_option("log", "file"):
            logf_templ = self.get("log", "file")
            curr_logf = pathparse.append_time_to_filename(logf_templ, ".log")
            self.__logfile_dir = os.path.dirname(curr_logf)

            if self.has_option("log", "level") :
                self.logger = log.LogFactory.getLogger(self.getProfileName(),
                                                   curr_logf,
                                                   self.getint("log", "level"))
            else :
                self.logger = log.LogFactory.getLogger(self.getProfileName(),
                                                   curr_logf)

            self.logger.info(_("Log output for [%(profile)s] is directed to file '%(file)s'.")\
                               % {'profile' : self.getProfileName(),
                                  'file': curr_logf})
        else:
        # if no file is specified, use the logger's default (no log file)
# TODO: Raise an assertion exception if no log section was found ?!
            self.logger = log.LogFactory.getLogger(self.getProfileName())
            self.logger.info(_("Log output for [%s] is not directed into a file.")\
                                % (self.getProfileName()))
        self.__current_logfile = curr_logf

    def __cleanup_logfiles(self):
        """Renames the unique log file that was created in the constructor
        and rotates existing files in order to keep the log directory
        clear.
        """
        _max_num = 6
        logf_src = self.get_current_logfile()
        if logf_src is not None:
            logf_target = self.__get_logfile_template()
            if local_file_utils.path_exists(logf_src):
                try:
                    local_file_utils.rename_errors_ignored(logf_target, "%s.0" % logf_target)
                    local_file_utils.compress_rotated_files(logf_target, _max_num)
                    local_file_utils.rename_rotating(logf_src, logf_target, _max_num)
                except OSError, error:
                    self.logger.exception(_("Unable to rename log file '%(src)s'->'%(dst)s': %(err)s")\
                                    % {'src': logf_src, 'dst': logf_target, 'err': str(error)})
        self.logger = None

    def read(self, filename = None):
        """Reads the configuration file and returns its content. Moreover it
        sets up a logger with appropriate log file and log level. This method
        overwrites the 'read' method from base class.
        
        @param filename: Full path of configuration file.
        @type filename:  String
        
        @return: The read configuration
        @rtype:  Same type as the base class returns
        """
        if filename:
            self.conffile = filename
        retValue = ConfigParser.ConfigParser.read(self, self.conffile)

        if len(retValue) == 0 :
            raise exceptions.SBException(_("The config file '%s' couldn't be read!")\
                                % self.conffile)
        return retValue

    def validateConfigFileOpts(self):
        self.logger.debug("Validating config file.")
        if (self.valid_options is None):
            return
        for section in self.sections():
            try:
                for key in self.options(section):
                    if (not self.valid_options.has_key(section)):
                        raise exceptions.NotValidSectionException (_("section [%(section)s] in '%(configfile)s' should not exist, aborting") % {'section': section, 'configfile' :self.conffile})
                    if key == 'stop_if_no_target':
                        self.logger.warning("Obsolete option `%s` in configuration found. Ignored." % key)
                    if (self.valid_options[section].has_key(key) or self.valid_options[section].has_key('*')):
                        continue
                    raise exceptions.NonValidOptionException ("key '%s' in section '%s' in file '%s' is not known, a typo possibly?" % (key, section, self.conffile))
            except exceptions.SBException, e:
                self.logger.error(str(e))
                raise e
        return True

    def allOpts(self):
        retVal = []
        for section in self.sections():
            for key in self.options(section):
                value = self.get(section, key, raw = True)
                retVal.append((key, value))
        return retVal

    def get_compress_format(self):
        if self.has_option("general", "format"):
            _compr = self.get("general", "format")
            if str(_compr) == "1":  # sbackup < 0.11 compatibility hack
                _compr = "gzip"
                self.set("general", "format", _compr)
        else:
            defaults = self.__get_default_config_obj()
            _compr = defaults.get_compress_format()
        return _compr


    def setSchedule(self, isCron, value):
        """Set the backup Schedule.
        
        @param isCron : schedule type (cron/anacron)
        @param value : a string containing the value to set for Cron/Anacron. \
                        Valid values for Anacron are: \
                            daily/monthly/hourly/weekly
                        Valid values for Cron:
                            cron expression to add at /etc/cron.d/sbackup.                            
        """
        _logger = log.LogFactory().getLogger()  # in some cases a logger is not yet available

        if isCron not in SCHEDULE_TYPES:
            raise exceptions.NonValidOptionException("Invalid schedule type given")

        if not self.has_section("schedule"):
            self.add_section("schedule")

        if isCron == SCHEDULE_TYPE_ANACRON:     # (simple scheduling)
            if value in _ANACRON_SERVICES :
                if self.has_option("schedule", "cron"):
                    _logger.debug("Removing Cron config to set Anacron config.")
                    self.remove_option("schedule", "cron")
                _logger.debug("Setting Anacron config to: %s" % value)
                self.set("schedule", "anacron", value)
            else:
                raise exceptions.NonValidOptionException("Valid values for anacron are: "\
                        "%s, got '%s' instead." % (str(_ANACRON_SERVICES), value))

        elif isCron == SCHEDULE_TYPE_CRON:  # Cron entry given (precise scheduling)
            if self.has_option("schedule", "anacron"):
                _logger.debug("Removing Anacron config to set Cron config.")
                self.remove_option("schedule", "anacron")
            _logger.debug("Setting cron config to: %s" % value)
            self.set("schedule", "cron", value)

    def get_schedule_and_probe(self):
        """Retrieves the current schedule state from configuration settings
        and from filesystem.

        @return: (isCron, value) a tuple where isCron = 0 if Anacron is used \
                 and isCron = 1 Cron is used. If no schedule setup has been \
                 found, 'None' is returned.
                 
        """
        _value = None
        if not self.has_section("schedule") \
                or (not self.has_option("schedule", "cron") \
                and not self.has_option("schedule", "anacron")):

            # no entry in configuration found, look at Cron/Anacron directly
            self.logger.info(_("No schedule defined in configuration file. Probing from filesystem."))
            # for anacron we check for link rather exists to cope with cases of broken links
            if os.path.islink("/etc/cron.hourly/sbackup"):
                self.logger.debug("Anacron hourly has been found")
                return (SCHEDULE_TYPE_ANACRON, "hourly")
            elif os.path.islink("/etc/cron.daily/sbackup"):
                self.logger.debug("Anacron daily has been found")
                return (SCHEDULE_TYPE_ANACRON, "daily")
            elif os.path.islink("/etc/cron.weekly/sbackup"):
                self.logger.debug("Anacron weekly has been found")
                return (SCHEDULE_TYPE_ANACRON, "weekly")
            elif os.path.islink("/etc/cron.monthly/sbackup"):
                self.logger.debug("Anacron monthly has been found")
                return (SCHEDULE_TYPE_ANACRON, "monthly")

            if os.path.exists("/etc/cron.d/sbackup"):
                _value = local_file_utils.readfile("/etc/cron.d/sbackup")
                self.logger.debug("Custom Cron has been found: %s" % _value)
                _value = parse_cronexpression(_value)
                return (SCHEDULE_TYPE_CRON, _value)
            # none has been found
            return None

        else:
            # scheduling is stored in configuration file
            if self.has_option("schedule", "cron"):
                _value = self.get("schedule", "cron")
                self.logger.debug("Schedule type Cron found in Config: %s" % _value)
                return (SCHEDULE_TYPE_CRON, _value)
            elif self.has_option("schedule", "anacron"):
                _value = self.get("schedule", "anacron")
                self.logger.debug("Schedule type Anacron found in Config: %s" % _value)
                return (SCHEDULE_TYPE_ANACRON, _value)
            else:
                return None

    def get_schedule(self):
        """Returns the scheduling information that are currently
        set in this configuration regardless of filesystem state.
        """
        # scheduling is stored in configuration file
        ret_val = None
        if self.has_option("schedule", "cron"):
            _value = self.get("schedule", "cron")
            self.logger.debug("Schedule type Cron found in Config: %s" % _value)
            ret_val = (1, _value)
        elif self.has_option("schedule", "anacron"):
            _value = self.get("schedule", "anacron")
            self.logger.debug("Schedule type Anacron found in Config: %s" % _value)
            ret_val = (0, _value)
        else:
            ret_val = None
        return ret_val

    def remove_schedule(self):
        """Removes all options stored in section 'schedule'. The section
        itself remains. Existing entries in crontab etc. are not touched.
        
        @return: Flag, whether something was removed or not. True is returned
                    in the case that any schedule information was found by
                    'getSchedule'.
                    
        The return of this flag is somewhat a hack to force the GUI to enable
        the Save button for the rare case that:
        * the user wants to remove the schedule (i.e. sets it to 'Never')
        * no schedule information is stored in the configuration file
        * a script is linked in Anacron directories or in 'etc/cron.d/' which
          is found when probing the filesystem for schedules.          
          
        """
        _something_removed = False
        if self.get_schedule_and_probe() is not None:
            _something_removed = True
            for _option in self.options("schedule"):
                self.logger.debug("Removing ('schedule','%s') "\
                                "from configuration." % _option)
                self.remove_option("schedule", _option)
        return _something_removed

    def getProfileName(self):
        """Returns the current profile name for the current ConfigManager.
        
        @return: the current profile name if the config file name match the
                 naming convention or Unknow otherwise
        @raise exceptions.SBException: if the configfile isn't set
        
        @todo: Implement Command-Query Separation Principle (CQS)!
        
        """
        if self.__profileName :
            return self.__profileName

        if not self.conffile:
            raise exceptions.SBException(_("The config file is not set yet into this ConfigManager"))
        self.__profileName = self.__conffile_hdl.get_profilename(self.conffile)
        return self.__profileName

    def is_default_profile(self):
        """Checks whether this configuration is the default configuration
        (i.e. default profile).
        
        @return: True if this is the default profile, False otherwise.
        
        """
        if not self.conffile:
            raise exceptions.SBException(_("The config file is not set yet into this ConfigManager"))
        is_default = is_default_profile(self.conffile)
        return is_default

    def getProfiles(self):
        """Returns the list of defined backup profiles. Both, active and
        disabled profiles are retrieved.
         
        @return: a dictionary of {profilename: [path_to_conffile, enabled]} 
        
        """
        profdir = self.__conffile_hdl.get_profilesdir(self.conffile)
        self.logger.debug("Getting profiles from `%s`" % profdir)
        _prfls = get_profiles(profdir)
        # debug output of found profiles
        for _prfk in _prfls.keys():
            self.logger.debug("Found profile '%s' (active = %s)" % (_prfk,
                                                            _prfls[_prfk][1]))
        # end of debug output
        return _prfls

    def set_logdir(self, logdir):
        """Sets the given directory as current log directory for use with
        log files.        
        """
        self.__logfile_dir = logdir

    def get_logdir(self):
        """Returns the currently set directory for log files.
        """
        return self.__logfile_dir

    def set_logfile_templ_to_config(self):
        """Retrieves the path to log file and writes it into the configuration.

        By 'template' is indicated this is the base logfile name. It is
        extended by the current time if a logfile is actually created.         
        """
        logf = self.__get_logfile_template()
        if not self.has_section("log") :
            self.add_section("log")
        self.set("log", "file", logf)

    def __get_logfile_template(self):
        """Builds the full path to log file for this configuration and
        returns it. The log file for the default profile is named
        'sbackup.log', log files for other profiles are extended by
        the profile's name to keep them unique and avoid problems while logging.
        
        By 'template' is indicated this is the base logfile name. It is
        extended by the current time if a logfile is actually created. 

        """
        logfname = get_logfile_name_template(self.conffile)
        logf = os.path.join(self.__logfile_dir, logfname)
        return logf

    def get_current_logfile(self):
        """Returns the full path of the log file currently used.
        Might return None.
        """
        return self.__current_logfile

    def __clear_report_section(self):
        """Any options present in the report section are removed. The section
        itself is not removed. 
        """
        _section = "report"
        self.remove_section(_section)
        if not self.has_section(_section):
            self.add_section(_section)

    def saveConf(self, configfile = None):
        """Saves the configuration (i.e. writes it into the specified file
        and sets scheduling in the case this is the default profile).
         
        @param configfile: The config file in which to write the configuration.
                            Default is in the default location
                             
        @todo: What happens if the config file is 'Save as...' in order to
                store a configuration aside but the scheduling is different
                from the *original* configuration that is used by sbackup?
        """
        if configfile is not None:
            fld = local_file_utils.openfile(configfile, True)
        else:
            fld = local_file_utils.openfile(self.conffile, True)
        self.write(fld)
        fld.close()
        if configfile is None and system.is_superuser():
            self.write_schedule()

    def write_schedule(self):
        """Write the schedule from the configuration file. Scheduling is only
        written for admin default profiles.
        
        """
        if not system.is_superuser():
            self.logger.warning("Schedules are not implemented for regular users")
            return

        if not self.is_default_profile():
            self.logger.warning("Schedules are not implemented for non-default profiles")
            return

        self.__erase_services()

        if not self.has_section("schedule"):
            return

        if self.has_option("schedule", "cron"):
            self.logger.debug("Writing Cron entry")
            _crpath = _CRON_PATH_TEMPLATE % (_CRON_SERVICE, _LAUNCHER_NAME_CRON)
            local_file_utils.writetofile(_crpath, self.__make_cronfile_content())

        elif self.has_option("schedule", "anacron"):
            _anacr = self.get("schedule", "anacron")
            self.logger.debug("Writing Anacron entry `%s`" % _anacr)
            assert _anacr in _ANACRON_SERVICES, "Anacron entry found in config is invalid"
            _anacrpath = _CRON_PATH_TEMPLATE % (_anacr, _LAUNCHER_NAME_CRON)
            if _anacr in _ANACRON_SERVICES:
                os.symlink(self.__servicefile, _anacrpath)
            else :
                self.logger.warning("Anacron entry `%s` is invalid" % _anacr)

    def __make_cronfile_content(self):
        """Collects required data in order to create the content for
        the file used to setup CRON (precise setting). 
        """
        _cronheader = ConfigManagerStaticData.get_cronheader()
        _cronexpr = self.get("schedule", "cron")
        assert (_cronexpr is not None) and (_cronexpr != "None"), \
               "Cron entry `%s` found in config is invalid" % _cronexpr
        _cronuser = self.__schedules_user
        _execline = "if [ -x '%s' ]; then %s; fi;" % (self.__servicefile,
                                                      self.__servicefile)
        # Note: cron requires a newline at the end of the file
        _content = "%s\n%s\t%s\t%s\n" % (_cronheader, _cronexpr,
                                         _cronuser, _execline)
        return _content

    def __erase_services(self):
        """Removes Cron and Anacron service from /etc/cron.*
        """
        for serv in _SCHEDULE_SERVICES:
            path = _CRON_PATH_TEMPLATE % (serv, _LAUNCHER_NAME_CRON)
            try:
                os.unlink(path)
            except OSError:
                pass
            else:
                print "Cron entry `%s` was removed" % path

    def testMail(self):
        """Test the mail settings
        
        @return: True if succeded
        @raise exceptions.SBException: the error message why it didn't run
        
        @todo: Not the right place for this kind of functionality?!?
        @todo: Implement specific `MailError` exception.
        """
        if not self.has_option("report", "to"):
            raise exceptions.SBException (_("No receiver set."))
        if not self.has_option("report", "smtpserver") :
            raise exceptions.SBException (_("No SMTP server set."))

        if (self.has_option("report", "smtpuser") and not self.has_option("report", "smtppassword")):
            raise exceptions.SBException (_("Username set but no password specified."))

        if (not self.has_option("report", "smtpuser") and self.has_option("report", "smtppassword")):
            raise exceptions.SBException (_("Password set but no username specified."))

        if not self.has_option("report", "smtptls") and (self.has_option("report", "smtpcert") or self.has_option("report", "smtpkey)")) :
            raise exceptions.SBException (_("A certificate and key file is given while SSL option (smtptls=1) is not set.\nSelect SSL in order to use Certificate and Key."))

        if self.has_option("report", "smtptls") and ((self.has_option("report", "smtpcert") and not self.has_option("report", "smtpkey")) \
            or (not self.has_option("report", "smtpcert") and self.has_option("report", "smtpkey"))):
            raise exceptions.SBException (_("When specifying a SSL certificate or key file, a key file resp. certificate is mandatory."))

        try :
            server = smtplib.SMTP()
            # getting the connection
            if self.has_option("report", "smtpport") :
                server.connect(self.get("report", "smtpserver"), self.get("report", "smtpport"))
            else :
                server.connect(self.get("report", "smtpserver"))

            if self.has_option("report", "smtptls") and self.get("report", "smtptls") == "1":
                if self.has_option("report", "smtpcert") and self.has_option("report", "smtpkey") :
                    server.starttls(self.get("report", "smtpkey"), self.get("report", "smtpcert"))
                else :
                    server.starttls()
            if self.has_option("report", "smtpuser") and self.has_option("report", "smtppassword") :
                server.login(self.get("report", "smtpuser"), self.get("report", "smtppassword"))

            server.helo()
            server.close()
            return True
        except Exception, e:
            raise exceptions.SBException(e)

    def isConfigEquals(self, config):
        """Compares this configuration and the given configuration.

        @param config : a configManager instance
        @return: True if the config are equals, False otherwise
        @rtype: boolean
        
        """
        if not (isinstance(config, ConfigManager) or config is None):
            raise exceptions.SBException("Can't compare a ConfigManager with type '%s'"\
                            % str(type(config)))
        if config is None:
            return False

        for s in self.sections() :
            if not config.has_section(s) :
                return False
            else :
                for o in self.options(s) :
                    if not config.has_option(s, o):
                        return False
                    else :
                        if type(self.get(s, o)) != type(config.get(s, o)) :
                            if type(self.get(s, o)) is not str :
                                self.set(s, o, repr(self.get(s, o)))
                            if type(config.get(s, o)) is not str :
                                config.set(s, o, repr(config.get(s, o)))
                        if not self.get(s, o) == config.get(s, o) :
                            return False
                for o in config.options(s) :
                    if not self.has_option(s, o):
                        return False
                    else :
                        if type(self.get(s, o)) != type(config.get(s, o)) :
                            if type(self.get(s, o)) is not str :
                                self.set(s, o, repr(self.get(s, o)))
                            if type(config.get(s, o)) is not str :
                                config.set(s, o, repr(config.get(s, o)))
                        if not self.get(s, o) == config.get(s, o) :
                            return False
        return True


class ConfigurationFileHandler(object):
    """Determines paths to configuration files and profiles
    under consideration of users privileges.
    
    When running in special `Developer mode` paths are read 
    """

    def __init__(self):
        self.__super_user = False
        self.__check_for_superuser()

    def __check_for_superuser(self):
        """Checks whether the application was invoked with super-user rights.
        If so, the member variable 'self.__super_user' is set.
        
        :todo: Here should no distinction between user/superuser be necessary!
        
        """
        if system.is_superuser():
            self.__super_user = True

    def get_conffile(self):
        """default profile config file is determined
        """
        if self.__super_user:
            conffile = os.path.join(ConfigManagerStaticData.get_superuser_confdir(),
                                    ConfigManagerStaticData.get_default_conffile())
        else:
            conffile = os.path.join(self.get_user_confdir(),
                                     ConfigManagerStaticData.get_default_conffile())
        return conffile

    def get_default_conffile_fullpath(self):
        """Returns the full path (incl. filename) o the default config file.
        """
        conffile = os.path.join(self.get_user_confdir(),
                                ConfigManagerStaticData.get_default_conffile())
        return conffile


    def get_profilesdir(self, conffile):
        """the config directory is determined from given conffile. It is
        supposed that the profile directory is placed in the same
        directory as the configuration file (i.e. given conf file is
        the default profile's file.
        
        """
        confdir = os.path.dirname(conffile)
        profdir = os.path.join(confdir, ConfigManagerStaticData.get_profiles_dir())
        return profdir

    def get_user_confdir(self):
        """Get the user config dir using the XDG specification.
        
        :todo: Is method `ConfigManagerStaticData.get_superuser_confdir()` still relevant?
        """
        if self.__super_user:
            confdir = local_file_utils.normpath("/etc")
        else:
            confdir = local_file_utils.normpath(system.get_user_config_dir(), "sbackup")
        if not local_file_utils.path_exists(confdir) :
            local_file_utils.makedirs(confdir)
        return confdir

    def get_user_datadir(self):
        """Get the user datas dir using the XDG specification.
        """
        datadir = local_file_utils.normpath(system.get_user_data_dir(), "sbackup")
        if not local_file_utils.path_exists(datadir) :
            local_file_utils.makedirs(datadir)
        return datadir

    def get_user_tempdir(self):
        """Returns the user's temporary directory. Currently always
        the directory `/tmp` is used.
        
        :return: full path to the SBackup tempdir
        
        :todo: Review the use of '/tmp' as tempdir as solution for several\
               different distributions?
        :todo: Put the definition of used paths into `ConfigManagerStaticData`!
               
        """
        if self.__super_user:
            tempdir = os.path.join("/tmp", "sbackup/")
        else:
            tempdir = os.path.join(self.get_user_datadir(), "tmp/")

        if not os.path.exists(tempdir) :
            os.mkdir(tempdir)

        #LP #785495: make the temp directory RWX only by owner
        if (os.stat(tempdir).st_mode & 0777) != 0700:
            os.chmod(tempdir,0700)
            
        return tempdir

    def get_profilename(self, conffile):
        # find the profile 
        cfile = os.path.basename(conffile)

        if cfile == ConfigManagerStaticData.get_default_conffile():
            profilename = ConfigManagerStaticData.get_default_profilename()
        else :
            m = ConfigManagerStaticData.get_profilename_re().match(cfile)
            if not m:
                profilename = ConfigManagerStaticData.get_unknown_profilename()
            else :
                profilename = m.group(1)

        return profilename


class ConfigManagerStaticData(object):
    """Any static data related to configurations are stored here.
    
    @todo: Refactor this class into a class containing a default
            configuration, one containing name definitions, and one
            containing path and file names.
            
    :todo: Simplify and use properties!
    
    """
    __cronheader = "SHELL=/bin/bash \n"\
        "PATH=/usr/local/sbin:/usr/local/bin:/sbin:/bin:/usr/sbin:/usr/bin\n"

    __schedule_script_file = util.get_resource_file("sbackup-launch")

    __default_profilename = _("Default Profile")
    __unknown_profilename = _("Unknown Profile")

    # these variables should be read-only
    __logfile_basename = "sbackup"
    __logfile_ext = "log"

    __superuser_confdir = "/etc"
    __user_confdir_template = ".config/sbackup"
    __default_config_file = "sbackup.conf"

    __profiles_dir = "sbackup.d"
    __profiles_dir_nssbackup = "nssbackup.d"

    __profilename_re = re.compile(r"^sbackup-(.+?).conf(-disable)?$")
    __profilename_nssbackup_re = re.compile(r"^nssbackup-(.+?).conf(-disable)?$")

    # configuration's existing sections and options
    __our_options = {
     'general'         : { 'mountdir'        : str,
                        'target'          : str ,
                           'lockfile'         : str ,
                           'maxincrement'     : int ,
                           'format'         : str,
                           'splitsize'     : int,
                           'purge'         : str,
                           'followlinks'     : int,
                           'stop_if_no_target' : int,
                        'packagecmd'    : str},
     'log'             : {'level' : int , 'file' : str },
     'report'         : {'from' :str, 'to' : str, 'smtpserver' : str,
                        'smtpport' : int, 'smtpuser' : str,
                        'smtppassword' : str, 'smtptls' : int,
                       'smtpcert': str, 'smtpkey': str },
     'dirconfig'    : { '*' : str },
     'exclude'         : { 'regex' : list, 'maxsize' : int },
     'places'         : { 'prefix' : str },
     'hooks'        : {'pre-backup': list, 'post-backup': list },
     'schedule'     : {'anacron' : str , 'cron' : str }
    }

    __loglevels = {    '10' : ("Debug", 0),
                    '20' : ("Info", 1),
                    '30' : ("Warning", 2),
                    '40' : ("Error", 3)}

    __simple_schedule_freqs = {    "hourly"    : 0,
                                    "daily"        : 1,
                                    "weekly"    : 2,
                                    "monthly"    : 3
                                }

    __cformats = ['none', 'gzip', 'bzip2']

    __splitsize = {    0        : _('Unlimited'),
                        100        : _('100 MiB'),
                        250        : _('250 MiB'),
                        650     : _('650 MiB'),
                        2048     : _('2 GiB (FAT16)'),
                        4096    : _('4 GiB (FAT32)'),
                        - 1        : _('Custom')
                    }
    __known_ftypes = {    "mp3"    : _("MP3 Music"),
                            "avi"    : _("AVI Video"),
                            "mpeg"    : _("MPEG Video"),
                            "mpg"    : _("MPEG Video"),
                            "mkv"    : _("Matrjoshka Video"),
                            "ogg"    : _("OGG Multimedia container"),
                            "iso"    : _("CD Images")
                        }

    def __init__(self):
        pass

    @classmethod
    def get_simple_schedule_frequencies(cls):
        """Returns a dictionary of schedule frequencies and IDs.
        """
        return cls.__simple_schedule_freqs

    @classmethod
    def get_cronheader(cls):
        """Returns the static cron header.
        """
        return cls.__cronheader

    @classmethod
    def get_schedule_script_file(cls):
        """Returns the service file (helper script) that is called when
        using scheduled backups.
        """
        return cls.__schedule_script_file

    @classmethod
    def get_default_logfile(cls):
        """Returns the name of the logfile for the default profile.        
        """
        logfname = "%s.%s" % (cls.__logfile_basename,
                              cls.__logfile_ext)
        return logfname

    @classmethod
    def get_valid_loglevels(cls):
        """Returns a dictionary of valid log levels (a number) and
        corresponding log level name.
        """
        return cls.__loglevels

    @classmethod
    def get_profile_logfile(cls, profilename):
        """Returns the name of the logfile for a certain profile
        named `profilename`.        
        """
        logfname = "%s-%s.%s" % (cls.__logfile_basename,
                                 profilename,
                                 cls.__logfile_ext)
        return logfname

    @classmethod
    def get_profiles_dir(cls):
        """Returns the name (only basename, no path) of the directory where
        profile configurations are stored.        
        """
        return cls.__profiles_dir

    @classmethod
    def get_profiles_dir_nssbackup(cls):
        """Returns the name (only basename, no path) of the directory where
        profile configurations are stored.        
        """
        return cls.__profiles_dir_nssbackup

    @classmethod
    def get_profilename_re(cls):
        return cls.__profilename_re

    @classmethod
    def get_profilename_nssbackup_re(cls):
        return cls.__profilename_nssbackup_re

    @classmethod
    def get_default_profilename(cls):
        return cls.__default_profilename

    @classmethod
    def get_unknown_profilename(cls):
        return cls.__unknown_profilename

    @classmethod
    def get_superuser_confdir(cls):
        return cls.__superuser_confdir

    @classmethod
    def get_user_confdir_template(cls):
        return cls.__user_confdir_template

    @classmethod
    def get_default_conffile(cls):
        return cls.__default_config_file

    @classmethod
    def get_our_options(cls):
        return cls.__our_options

    @classmethod
    def get_splitsize_dict(cls):
        return cls.__splitsize

    @classmethod
    def get_known_ftypes_dict(cls):
        return cls.__known_ftypes

    @classmethod
    def get_compr_formats(cls):
        return cls.__cformats


class Configuration(object):
    """Encapsulates a configuration, that is a set of backup profile related
    settings.

    Configuration values are stored in their natural type and format instead
    of the type and format that is used for storage of the configuration
    (i.e. Boolean values get the value True/False, not '0/1' or 'yes/no').
    The transformation of the actual configuration values to data that is
    written to disk is done in the ConfigManager.
    
    @note: This is an early attempt to refactor the configuration thing:
            * there are objects that represent the actual profile settings
            * there are certain global preferences that are valid for
                each profile (e.g. prefix, mountdir, etc.)
            * there is a Manager class that is responsible for handling
                of several profile configurations
            * there is an object that forms an interface between the actual
                configuration and their storage on disk, reading, evaluating
                and so on (probably derived from ConfigParser)
    """

    def __init__(self):
        self._maxinc = 0
        self._cformat = ""
        self._follow_links = False
        self._splitsize = 0

        self._dirconf = {}
        self._regex_excludes = ""
        self._max_filesize = 0

        self._purge = ""
        self._target = ""

        # administrative settings
        self._mountdir = ""
        self._lockfile = ""
        self._prefix = ""

        self._logdir = ""
        self._loglevel = 0

# TODO: Collect report related settings in separate class.
        self._report_smtpfrom = ""

        self._schedule = [False, ""]    # isCron = False, frequency = ""

    # access methods
    def get_report_smtpfrom(self):
        return self._report_smtpfrom

    def get_schedule(self):
        return self._schedule

    def get_regex_excludes(self):
        return self._regex_excludes

    def get_max_filesize(self):
        return self._max_filesize

    def get_max_increment(self):
        return self._maxinc

    def get_compress_format(self):
        return self._cformat

    def get_follow_links(self):
        if not isinstance(self._follow_links, types.BooleanType):
            raise TypeError("Boolean value for 'followlinks' expected. "\
                            "Found '%s' instead." % type(self._follow_links))
        return self._follow_links

    def get_split_size(self):
        return self._splitsize

    def get_dir_config(self):
        return self._dirconf

    def get_mountdir(self):
        return self._mountdir

    def get_purge(self):
        return self._purge

    def get_target(self):
        return self._target

    def get_lockfile(self):
        return self._lockfile

    def get_prefix(self):
        return self._prefix

    def get_logdir(self):
        return self._logdir

    def get_loglevel(self):
        return self._loglevel


class _DefaultConfiguration(Configuration):
    """Abstract base class for a default configuration. A default
    configuration is basically a configuration however the settings
    are predefined (and read-only in the best case).
    
    @todo: Change this: the default configuration should not be a derived
            class rather an instance of a regular configuration which gets
            its content from a default config file (in share). Doing so,
            would be more flexible? What is with /home/$username?
    
    """

    def __init__(self):
        Configuration.__init__(self)

        self._regex_excludes = "^/home/[^/]+?/\.gvfs(/|$),"\
                               "^/home/[^/]+?/\.thumbnails(/|$),"\
                               "^/home/[^/]+?/\..+/[tT]rash(/|$),"\
                               "^/home/[^/]+?/\.cache(/|$),"\
                               "^/home/[^/]+?/\..+/[cC]ache(/|$),"\
                               "^/home/[^/]+?/\..+/lock(/|$),"\
                               "~$"
        self._maxinc = 7
        self._cformat = 'none'
        self._follow_links = False
        self._splitsize = 0
        self._max_filesize = -1 #Bugfix LP #146618

        self._loglevel = "20"  # set values as strings (LP #1159705)
        self._report_smtpfrom = Infos.SMTPFROM

        self._lockfile = "/var/lock/sbackup/sbackup.lock"
        self._prefix = '/usr'
        self._purge = "30"


class _DefaultConfigurationForAdmins(_DefaultConfiguration):
    """Derived default configuration that specializes it for
    privileged users. Only settings that are unique for
    admins are defined here.
    
    @todo: Change this: the default configuration should not be a derived
            class rather an instance of a regular configuration which gets
            its content from a default config file (in share). Doing so,
            would be more flexible? What is with /home/$username?
    
    """

    def __init__(self):
        _DefaultConfiguration.__init__(self)

        self._mountdir = "/mnt/sbackup"
        self._target = "/var/backup"

        self._dirconf = { '/etc/'            : '1',
                            '/var/'            : '1',
                            '/home/'        : '1',
                            '/var/cache/'    : '0',
                            '/var/tmp/'        : '0',
                            '/var/spool/'    : '0',
                            '/var/run/'        : '0',
                            '/usr/local/'    : '1',
                            '/media/'        : '0' }

        self._logdir = "/var/log/sbackup"

        self._schedule = [False, "daily"]


class _DefaultConfigurationForUsers(_DefaultConfiguration):
    """Derived default configuration that specializes it for
    normal users. Only settings that are unique for normal users
    are defined here.
    
    @todo: Change this: the default configuration should not be a derived
            class rather an instance of a regular configuration which gets
            its content from a default config file (in share). Doing so,
            would be more flexible? What is with /home/$username?
    
    """

    def __init__(self):
        _DefaultConfiguration.__init__(self)

        _hdl = ConfigurationFileHandler()
        self._mountdir = os.path.join(_hdl.get_user_datadir(), "mountdir")
        self._target = os.path.join(_hdl.get_user_datadir(), "backups")

        self._dirconf = { "%s%s" % (system.get_user_home_dir(), os.sep) : '1' }

        self._logdir = os.path.join(_hdl.get_user_datadir(), "log")

        self._schedule = [False, ""]    # no scheduling



#   Simple Backup - definition of custom exceptions
#
#   Copyright (c)2008-2010,2013: Jean-Peer Lorenz <peer.loz@gmx.net>
#   Copyright (c)2007-2008: Ouattara Oumar Aziz <wattazoum@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
"""
:mod:`exceptions` -- custom exceptions
======================================

.. module:: util.exceptions
   :synopsis: Defines custom exceptions
.. moduleauthor:: Ouattara Oumar Aziz (alias wattazoum) <wattazoum@gmail.com>
.. moduleauthor:: Jean-Peer Lorenz <peer.loz@gmx.net>

"""


class ErrorDescription(object):
    def __init__(self, errorcode, errorname, errormessage):
        self.code = errorcode
        self.name = errorname
        self.message = errormessage

    def __str__(self):
        res = "code: %s  name: %s  message: %s" % (self.code, self.name, self.message)
        return res


class SigTerminatedError(Exception) :
    """This Exception is thrown if the backup process receives a
    SIGTERM/SIGKILL signal.
    """

class SBException(Exception):
    """This class will help us distinguish Exception that we must
    handle (exceptions created by ourself) and exceptions due to 
    programming errors (Python Exceptions).
    """

SBackupError = SBException


class ResourceFileNotFoundError(SBackupError) :
    """This Exception is thrown if another application instance is
    already running.
    """

class InstanceRunningError(SBException) :
    """This Exception is thrown if another application instance is
    already running.
    """

class ApplicationLockError(SBException) :
    """This Exception is thrown if another application instance is
    already running.
    """

class BackupCanceledError(SBException) :
    """This Exception is thrown if the backup process is canceled from
    the indicator gui.
    """
    
class HookedScriptError(SBException) :
    """This Exception is thrown if a script run pre/post backup returns
    exitcodes != 0.
    """

class NotValidSnapshotException(SBException) :
    """
    This Exception is thrown by Snapshot validation.
    """

class NotValidSnapshotNameException(NotValidSnapshotException):
    """Exception launched when the name of a snapshot is not valid
    
    """


class NotValidSectionException(SBException) :
    """This Exception is thrown by Config Section validation.
    
    """

class NonValidOptionException(SBException):
    """Thrown when a config option is not Valid

    """

class CorruptedSBdictException(SBException):
    """Thrown when a SBdict is corrupted

    """

class TimeoutError(SBException):
    """Thrown when an IO operation times out.
    """

class RemoveSnapshotHasChildsError(SBackupError):
    """Is raised if snapshot removal fails.
    """

class CopyFileAttributesError(SBException):
    """Thrown when a destination does not support file modes (e.g. some
    ftp servers or FAT filesystems).
    
    """

class NotSupportedError(Exception):
    """Thrown when trying to call a stub.

    """

class NotifyException(SBException) :
    """This Exception is thrown by notifiers and listeners.

    """

class DBusException(SBException) :
    """This Exception is thrown when problems with DBus occurs.

    """

class FileAccessException(SBException) :
    """This Exception is thrown when problems with remote paths occur.

    """

class FuseFAMException(FileAccessException):
    """Thrown when a Fuse mount fails

    """

class RemoteMountFailedError(FileAccessException):
    """This Exception is thrown when remote path is not mountable.

    """

class RemoteMountTestFailedError(FileAccessException):
    """This Exception is thrown when remote path is not mountable.

    """

class RemoteUmountFailedError(FileAccessException):
    """This Exception is thrown when remote path is not mountable.

    """

class FileAlreadyClosedError(SBException) :
    """This Exception is thrown when a file or stream cannot be closed because it is already closed.

    """



#   Simple Backup - snapshot handling
#
#   Copyright (c)2008-2010,2013: Jean-Peer Lorenz <peer.loz@gmx.net>
#   Copyright (c)2007-2008: Ouattara Oumar Aziz <wattazoum@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
"""
:mod:`SnapshotManager` --- Snapshot handler class
====================================================================

.. module:: SnapshotManager
   :synopsis: Defines a snapshot handler class
.. moduleauthor:: Jean-Peer Lorenz <peer.loz@gmx.net>
.. moduleauthor:: Ouattara Oumar Aziz (alias wattazoum) <wattazoum@gmail.com>

"""


from gettext import gettext as _
import traceback
import datetime
import time
import types

from sbackup.pkginfo import Infos
from sbackup.fs_backend import fam

from sbackup.ar_backend import tar
from sbackup.ar_backend.tar import SnapshotFile
from sbackup.ar_backend.tar import Dumpdir
from sbackup.ar_backend.tar import SnapshotFileWrapper
from sbackup.ar_backend.tar import ProcSnapshotFile
from sbackup.ar_backend.tar import get_dumpdir_from_list

import sbackup.util as Util
from sbackup.util import constants
from sbackup.core.snapshot import Snapshot
from sbackup.util.log import LogFactory
from sbackup.util.exceptions import SBException
from sbackup.util.exceptions import NotValidSnapshotException
from sbackup.util.exceptions import NotValidSnapshotNameException
from sbackup.util.exceptions import RemoveSnapshotHasChildsError
from sbackup.util.exceptions import NotSupportedError
from sbackup.util.exceptions import FileAccessException


_EXT_CORRUPT_SNP = ".corrupt"


class SnapshotManager(object):
    """Class responsible for handling and managing of several snapshots.
     
    :todo: Remove instance variables 'status' and implement an observer\
           pattern or progress function hooks! 
    
    :note: Rebasing of snapshots is disabled due to severe performance issues.
    """

    def __init__(self, destination):
        """Default constructor. Takes the path to the target backup
        directory as parameter.
        """
        if not isinstance(destination, types.StringTypes):
            raise TypeError("Destination path of type string expected. Got %s instead"\
                            % type(destination))

        self.logger = LogFactory.getLogger()

        self._fop = fam.get_file_operations_facade_instance()
        # This is the current directory used by this SnapshotManager
        self.__dest_path = destination

        # The list of the snapshots is stored the first time it's used,
        # so we don't have to re-get it later
        self.__snapshots = None

        # helper variables for displaying status messages
        self.statusMessage = None
        self.substatusMessage = None
        self.statusNumber = None

    def getStatus(self):
        """
        :return: [statusNumber,statusMessage,substatusMessage]
        
        :todo: Remove/refactor this!

        """
        return [self.statusNumber, self.statusMessage, self.substatusMessage]

    def get_snapshot_allformats(self, name):
        """Returns a certain snapshot, specified by its name, from the stored
        snapshots. If the snapshot could not be found, an exception is raised.
        
        :param name: the snapshot that is to be returned
        
        """
        for snp in self.get_snapshots_allformats() :
            if snp.getName() == name :
                return snp
        raise SBException(_("Snapshot '%s' not found ") % name)

    def get_snapshots_allformats_by_timespan_ro(self, from_date, to_date, force_reload = False):
        """Read-only variant that returns a list with *all* found snapshots, according to the
        given parameters. All versions of snapshots were returned.
        No snapshot are being removed or renamed.
        """
        snapshots = []
        for snp in self.get_snapshots_allformats_ro(force_reload):
            if from_date <= snp.getName()[:10] <= to_date:
                snapshots.append(snp)
        snapshots.sort(key = Snapshot.getName, reverse = True)
        return snapshots

    def get_snapshots_allformats_by_date_ro(self, by_date, force_reload = False):
        snapshots = []
        for snp in self.get_snapshots_allformats_ro(force_reload):
            if snp.getName().startswith(by_date):
                snapshots.append(snp)
        snapshots.sort(key = Snapshot.getName, reverse = True)
        return snapshots

    def get_snapshots_allformats_ro(self, force_reload = False):
        if self.__snapshots and not force_reload:
            pass
        else:
            self._read_snps_from_disk_allformats(read_only = True)
        snapshots = self.__snapshots
        snapshots.sort(key = Snapshot.getName, reverse = True)
        return snapshots

    def get_snapshots_allformats(self, fromDate = None, toDate = None, byDate = None,
                                 forceReload = False):
        """Returns a list with *all* found snapshots, according to the
        given parameters. All versions of snapshots were returned. The
        list is sorted from the latest snapshot to the earliest:
        
        - index 0  --- most recent snapshot 
        - index -1 --- oldest snapshot.
        
        :param fromDate: eg. 2007-02-17
        :param toDate:  2007-02-17
        :param byDate: 2007-02-17
        :param forceReload: True or false
        :return: list of snapshots 
        
        :todo: Re-factor this method using the CQS pattern and by simplifying!
                (e.g. method `reload` + `get...`
        :todo: Separate into 'get_snapshots( force_reload=False )',\
               'get_snapshots_by_timespan' and 'get_snapshot_by_date'!
        :todo: Clarify whether to rename or to delete corrupt snapshots!
               
        """
        snapshots = []    # list of found snapshots

        if fromDate and toDate:
            for snp in self.get_snapshots_allformats():
                if fromDate <= snp.getName()[:10] <= toDate:
                    snapshots.append(snp)
        elif byDate:
            for snp in self.get_snapshots_allformats():
                if snp.getName().startswith(byDate):
                    snapshots.append(snp)
        else:
            if self.__snapshots and not forceReload:
                pass
            else:
                self._read_snps_from_disk_allformats(read_only = False)
            snapshots = self.__snapshots

        return snapshots

    def _read_snps_from_disk_allformats(self, read_only = True):
        """Reads snapshots from the defined/set target directory and
        stores them in according class attribute.
        
        Unreadable snapshots are being renamed.
        """
        self.__snapshots = []
        listing = self._fop.listdir_fullpath(self.__dest_path)

        for _snppath in listing :
            _snpname = self._fop.get_basename(_snppath)
            try:
                self._fop.test_dir_access(_snppath)
            except FileAccessException, error:
                self.logger.info("Unable to access `%s'. Skipped." % _snppath)
                continue

            if _snpname.endswith(_EXT_CORRUPT_SNP):
                self.logger.info("Corrupt snapshot `%s` found. Skipped." % _snpname)
                continue
            try:
                self.__snapshots.append(Snapshot(_snppath))
            except NotValidSnapshotException, error :
                if isinstance(error, NotValidSnapshotNameException) :
                    self.logger.info(_("Invalid snapshot `%(name)s` found: Name of snapshot not valid.")\
                                        % { 'name': str(_snpname) })
                else: # rename only if name was valid but snapshot was invalid
                    self.logger.info(_("Invalid snapshot `%(name)s` found: %(error_cause)s.")\
                                        % { 'name': str(_snpname), 'error_cause' :error })
                    if not read_only:
                        self.logger.info("Invalid snapshot `%s` is being renamed." % _snpname)
                        if _snppath.endswith(".inc") or _snppath.endswith(".ful"):
                            _ren_snppath = "%s%s" % (_snppath[:-4], _EXT_CORRUPT_SNP)
                        else:
                            _ren_snppath = "%s%s" % (_snppath, _EXT_CORRUPT_SNP)
                        self._fop.rename(_snppath, _ren_snppath)

        self.__snapshots.sort(key = Snapshot.getName, reverse = True)

    def get_snapshots(self, fromDate = None, toDate = None, byDate = None,
                      forceReload = False):
        """Returns a list with found snapshots that matches the current
        snapshot format, according to the given parameters. The list is
        sorted from the latest snapshot to the earliest:
        
        - index 0  --- most recent snapshot 
        - index -1 --- oldest snapshot.
        
        :param fromDate: eg. 2007-02-17
        :param toDate:  2007-02-17
        :param byDate: 2007-02-17
        :param forceReload: True or false
        :return: list of snapshots
        
        """
        snps = []
        snps_all = self.get_snapshots_allformats(fromDate, toDate, byDate,
                                                  forceReload)
        for csnp in snps_all:
            if csnp.getVersion() == Infos.SNPCURVERSION:
                snps.append(csnp)
        self.__snapshots = snps
        # debugging output
        if self.logger.isEnabledFor(5):
            self.logger.debug("[Snapshots Listing - current format]")
            for csnp in snps:
                self.logger.debug(str(csnp))
        ###
        return snps

    def _copy_empty_snar(self, snp_source, copydest):
        """Creates an empty SnapshotInfo-file with the name 'copydest'
        from the SnapshotInfo-file contained in given source snapshot. Empty
        means, that no content but the header is copied.
        
        @todo: Review the self-creation of header in the case no was found.
                Is this necessary at all?
        
        """
        self.logger.debug("Create temporary SNARFILE to prepare merging")
        if not isinstance(snp_source, Snapshot):
            raise TypeError("Given parameter 'snp_source' must be of Snapshot "\
                        "type! Got %s instead." % type(snp_source))
        if not isinstance(copydest, str):
            raise TypeError("Given parameter 'copydest' must be of string "\
                        "type! Got %s instead." % type(copydest))

        # create a temporary snar file for merge result 
        _tmpfinal = copydest
        # get snar header from current snapshots
#XXX: Why not use getHeader here???
        _snarf = open(snp_source.getSnarFile())
        _header = _snarf.readline()
        if len(_header) > 0:
            # the SNAR file isn't empty
            sepcnt = 0
            while sepcnt < 2:
                readchar = _snarf.read(1)
                if len(readchar) != 1:
                    _snarf.close()
                    raise SBException(_("The snarfile header is incomplete !"))
                if readchar == '\0':
                    sepcnt += 1
                _header += readchar
            _snarf.close()
            self.logger.debug("Current SNAR Header (NULL replaced by newline):"\
                              "\n%s" % (_header.replace("\x00", "\n")))
        else:
            # the SNAR file is empty
            self.logger.debug("SNAR file empty, create the header manually")
            _snarf.close()
            _date = snp_source.getDate()
            _datet = datetime.datetime(_date['year'], _date['month'],
                                       _date['day'], _date['hour'],
                                       _date['minute'], _date['second'])


        # create temporary SNAR file and copy the retrieved header into it
        finalsnar = ProcSnapshotFile(SnapshotFile(_tmpfinal, True))

        if _header:
            snpif = open(_tmpfinal, 'w')
            snpif.write(_header)
            snpif.close()
        else :
            finalsnar.setHeader(_datet)
            _header = finalsnar.getHeader()
        return finalsnar

    def _merge_snarfiles(self, target_snpfinfo, target_excludes,
                               src_snpfinfo, res_snpfinfo):
        """Covers all actions for merging 2 given snar files into a single
        one. This is quite TAR specific - think it over where to place it!
        
        :Parameters:
        - `target_snpfinfo`: the resulting snapshot
        - `target_excludes`: set of the excludes file list of resulting snapshot
        - `src_snpfinfo`: the snapshot that should be merged into the target
        - `res_snpfinfo`: the name of the resulting SNAR file  
        
        The method returns a list containing files that needs to be extracted
        from the archive that was merged in. 
        
        :todo: Do we need to consider the order of the snar files?
        :todo: Needs more refactoring! (CQS)
        
        """
        self.logger.info("Merging SNARFILEs to make the transfer")

        if not isinstance(target_snpfinfo, SnapshotFileWrapper):
            raise TypeError("Given parameter 'target_snpfinfo' must be of "\
                        "SnapshotFileWrapper "\
                        "type! Got %s instead." % type(target_snpfinfo))
        if not isinstance(target_excludes, set):
            raise TypeError("Given parameter 'target_excludes' must be of "\
                        "type Set! "\
                        "Got %s instead." % type(target_excludes))
        if not isinstance(src_snpfinfo, SnapshotFileWrapper):
            raise TypeError("Given parameter 'src_snpfinfo' must be of "\
                        "SnapshotFileWrapper "\
                        "type! Got %s instead." % type(src_snpfinfo))
        if not isinstance(res_snpfinfo, SnapshotFileWrapper):
            raise TypeError("Given parameter 'res_snpfinfo' must be of "\
                        "SnapshotFileWrapper "\
                        "type! Got %s instead." % type(res_snpfinfo))

#        print "Parent (base) snar file:\n%s" % src_snpfinfo
        # list for storage of files that need to be extracted from merge source
        files_to_extract = []

        for target_record in target_snpfinfo.iterRecords():
            _tmp_dumpdirs = []
#TODO: A similar method to getContent would be nice!
            _curdir = target_record[SnapshotFile.REC_DIRNAME]
            # get the content (dumpdir entries) for current directory
            _curcontent = target_snpfinfo.getContent(_curdir)
            for _dumpdir in _curcontent:
#                print "\n  now processing dumpdir: %s" % _dumpdir
                _ctrl = _dumpdir.getControl()
                _filen = _dumpdir.getFilename()
                _ddir_final = None
                _was_excluded = False
                if _ctrl == Dumpdir.UNCHANGED:
                    # Item was explicitly excluded and is therefore not included in child
#                    _filenfull = os.path.join(_curdir, _filen)
#                    print "Full path: %s" % (_filenfull)
                    if self._fop.joinpath(_curdir, _filen) in target_excludes:
                        self.logger.debug("Path '%s' was excluded. Not merged." % _filen)
                        _was_excluded = True
                    else:
                        # Item has not changed and is therefore not included in child (i.e. target) snapshot.
                        # look for the item in the parent (i.e. base/source) snapshot
                        _basedumpd = get_dumpdir_from_list(\
                                                src_snpfinfo.getContent(_curdir),
                                                _filen)
                        _base_ctrl = _basedumpd.getControl()

                        if _base_ctrl == Dumpdir.UNCHANGED:
                            _ddir_final = _dumpdir

                        elif _base_ctrl == Dumpdir.INCLUDED:
                            _ddir_final = _basedumpd
                            files_to_extract.append(self._fop.joinpath(_curdir,
                                                                 _filen))
                        else:
                            raise SBException("Found unexpected control code "\
                                              "('%s') in snapshot file '%s'."\
                                              % (_ctrl, target_snpfinfo.get_snapfile_path()))

                elif _ctrl == Dumpdir.DIRECTORY:
                    _ddir_final = _dumpdir

                elif _ctrl == Dumpdir.INCLUDED:
                    _ddir_final = _dumpdir
                else:
                    raise SBException("Found unexpected control code "\
                                      "('%s') in snapshot file '%s'."\
                                      % (_ctrl, target_snpfinfo.get_snpfile_Path()))

                if not _was_excluded:
                    _tmp_dumpdirs.append(_ddir_final)
            # end of loop over dumpdirs 
            _final_record = target_record[:SnapshotFile.REC_CONTENT]
            _final_record.append(_tmp_dumpdirs)
            # write to the SnarFile
            res_snpfinfo.addRecord(_final_record)
        return files_to_extract

    def __makeSnpFull(self, snapshot):
        """Make an inc snapshot to a full one.
                
        :param snapshot: the snapshot to be converted
        :type snapshot: `Snapshot`
        :return: the new full snapshot
        :rtype: Snapshot
                
        :todo: Is it really neccessary to create a new snapshot or is it enough to call `setPath`?
                -> It is necessary since paths to e.g. snar file have changed.
        
        :postcondition: The snapshot has the same childs as before.
        
        """
        if snapshot.isfull():
            self.logger.info(_("Snapshot '%s' is already Full, nothing to do (not changing it to full).") % snapshot.getName())
            res_snp = snapshot
        else:
            childs = self._retrieve_childsnps(snapshot)

            if childs:
                fulname = snapshot.getName()[:-3] + 'ful'
                for _snp in childs:
                    _snp.setBase(fulname)
                    _snp.commitbasefile()

            path = snapshot.getPath()
            self._fop.rename(self._fop.joinpath(path, 'base'), self._fop.joinpath(path, 'base.old'))
            self._fop.rename(path, path[:-3] + 'ful')
            res_snp = Snapshot(path[:-3] + 'ful')

            # post-condition check
            # all childs are preserved
            postcond_child_names = self._retrieve_childsnps_names(res_snp)

            if len(childs) != len(postcond_child_names):
                raise AssertionError("Renaming of base of child snapshots was "\
                                     "not successful.")
            for _chl in childs:
                if _chl.getName() not in postcond_child_names:
                    raise AssertionError("Renaming of base of child snapshots "\
                                         "was not successful.")

        return res_snp

    def _retrieve_childsnps(self, snapshot):
        """Retrieves all snapshots that rely on the given parent
        `snapshot` and returns a list containing all child snapshots.
        
        """
        listing = self.get_snapshots(forceReload = False)
        child_snps = []
        for snp in listing :
            if snp.getBase() == snapshot.getName() :
                child_snps.append(snp)
        return child_snps

    def _retrieve_childsnps_names(self, snapshot):
        """Retrieves names of all snapshots that rely on the given parent
        `snapshot` and returns a list containing all child snapshot names.
        
        """
        listing = self._retrieve_childsnps(snapshot)
        child_snps = []
        for snp in listing:
            child_snps.append(snp.getName())
        return child_snps

    def _remove_standalone_snapshot(self, snapshot):
        if not self.is_standalone_snapshot(snapshot):
            raise RemoveSnapshotHasChildsError("The given snapshot '%s' is not stand-alone." % snapshot)
        self.logger.info("Removing '%s'" % snapshot.getName())
        self._fop.delete(snapshot.getPath())
        self.get_snapshots(forceReload = True)

    def is_standalone_snapshot(self, snapshot):
        _res = False
        _childs = self._retrieve_childsnps(snapshot = snapshot)
        if len(_childs) == 0:
            _res = True
        return _res

    def removeSnapshot(self, snapshot):
        """Public method that removes a given snapshot safely. The removal
        of a snapshot is more complicated than just to remove the snapshot
        directory since a snapshots could be the base of other snapshots.
        
        :param snapshot: the snapshot to be removed
        :type snapshot: `Snapshot`
        
        :note; Currently removal of freestanding snapshots is supported only.
        """
        self._remove_standalone_snapshot(snapshot)

    def remove_snapshot_forced(self, snapshot):
        """Removes snapshot directory forcefully.
        """
        self.logger.debug("Removing '%s'" % snapshot.getName())
        self._fop.delete(snapshot.getPath())
        self.get_snapshots(forceReload = True)

    def compareSnapshots(self, snap1, snap2):
        """Compare 2 snapshots and return and SBdict with the
        differences between their files. The format is
        {"file" : ("propsnap1|propsnap2",sonsbdict)}.
        
        """
        raise NotSupportedError

    def getSnpHistory(self, snapshot):
        """
        gets the list of preceding snapshots till the last full one
        :param snapshot : the given snapshot
        :return: a list of Snapshots starting from the most recent one to the full one
        :note: you'll need to reverse this list to make a revert 
        """
        if not snapshot :
            raise SBException("Please provide a snapshot to process")

        result = []
        # add the first snapshot
        result.append(snapshot)
        current = snapshot
        while (current.getBaseSnapshot()) :
            current = current.getBaseSnapshot()
            result.append(current)

        # Just for DEBUG
        if self.logger.isEnabledFor(10) :
            # get the history 
            history = "\n[%s history]" % snapshot.getName()
            for snp in result :
                history += "\n- %s" % snp.getName()
            self.logger.debug(history)

        return result

    def purge(self, purge, no_purge_snp):
        """Public method that processes purging of archive directory.
        
        :param mode: for the moment, only "log" and "simple" are supported
        :param no_purge: name of snapshot not being purged 
        
        :todo: We should try to remove the snapshots from fresh to old to avoid multiple re-base operations!
        
        """
        self.get_snapshots(forceReload = True)
        if purge == "log":
            self._do_log_purge(no_purge_snp)
        else:
            self._do_cutoff_purge(purge, no_purge_snp)
        self.get_snapshots(forceReload = True)

    def _do_log_purge(self, no_purge_snp = ""):
        """Logarithmic purge
        Keep progressivelly less backups into the past:
        Keep all backups from yesterday
        Keep one backup per day from last week.
        Keep one backup per week from last month.
        Keep one backup per month from last year.
        Keep one backup per quarter from 2nd last year.
        Keep one backup per year further in past.        
        """

        self.logger.info("Logarithmic purging")
        # compute years since begin of epoch: we need to go back this far
        _years_epoch = int(time.time() / (constants.SECONDS_IN_DAY * constants.DAYS_IN_YEAR))

        purge_plan = [ { "title" : "Last week", "nperiod" : 7,
                         "interval" : 1 },
                       { "title" : "Last month", "nperiod" : 3,
                         "interval" : constants.DAYS_IN_WEEK },
                       { "title" : "Last year", "nperiod" : 11,
                         "interval" : constants.DAYS_IN_MONTH },
                       { "title" : "2nd last year", "nperiod" : 4,
                         "interval" : constants.DAYS_IN_QUARTER },
                       { "title" : "remaining years", "nperiod" : _years_epoch,
                         "interval" : constants.DAYS_IN_YEAR }
                     ]

        _max_age = 2    # start value
        for pent in purge_plan:
            self.logger.info("Logarithm Purging [%s]" % pent["title"])
            _max_age = self.__purge_period(start = (_max_age - 1), nperiod = pent["nperiod"],
                                           interval = pent["interval"], no_purge_snp = no_purge_snp)

    def __purge_period(self, start, nperiod, interval, no_purge_snp):
        """period is given as `start` age and interval length in days.
        The period is repeated `nperiod` times.
        Within these timespans the defined number of backups must remain.
        """
        _number_to_keep = 1
        for j in range(0, nperiod):
            _min_age = start + (j * interval)
            _max_age = _min_age + (interval + 1)
            self._do_purge_in_timespan(_min_age, _max_age, _number_to_keep, no_purge_snp)
        return _max_age

    def _do_purge_in_timespan(self, min_age, max_age, number_to_keep, no_purge_snp = ""):
        """Simple purging is processed: all snapshots in timespan (i.e. younger
        than `max_age` and older than `min_age` are removed. Only freestanding
        snapshots are removed.
        Given snapshot `no_purge_snp` is never removed.
        The removal is terminated if `number_to_keep` snapshots remain.
        """
        _min_age = int(round(min_age))
        _max_age = int(round(max_age))
        assert _max_age > _min_age, "Given parameter max. age should be greater than min. age"

        if _min_age > 0:
            self.logger.debug("Purge in timespan\nRemove freestanding snapshots younger "\
                              "than %(max_age)s and older than %(min_age)s days."\
                              % {"max_age" : _max_age, "min_age" : _min_age})

            while True:
                _was_removed = False
                snapshots = self.get_snapshots()    # sort order: idx 0 = most recent
                snapshots = _get_snapshots_younger_than(snapshots, _max_age)
                snapshots = _get_snapshots_older_than(snapshots, _min_age)

                # debugging output
                if self.logger.isEnabledFor(5):
                    self.logger.debug("Snapshots in timespan - re-sorted]")
                    for csnp in snapshots:
                        self.logger.debug(str(csnp))
                ###

                _nsnps = len(snapshots)
                _maxidx = _nsnps - number_to_keep    # biggest valid index
                if _nsnps <= number_to_keep:
                    break
                for _idx in range(_maxidx):
                    snp = snapshots[_idx]
                    if snp.getName() == no_purge_snp:
                        self.logger.debug("`%s` skipped.")
                        continue

                    self.logger.debug("Checking '%s' for childs." % (snp))
                    childs = self._retrieve_childsnps(snapshot = snp)
                    if len(childs) == 0:
                        self.logger.debug("Snapshot '%s' has no childs "\
                                        "-> is being removed." % (snp))
                        self.remove_snapshot_forced(snapshot = snp) # it's freestanding
                        _was_removed = True
                        break
                if _was_removed is False:
                    break

    def _do_cutoff_purge(self, purge, no_purge_snp = ""):
        """Simple cut-off purging is processed: all snapshots older than
        a certain value are removed. During removal of snapshots the
        snapshot state (full, inc) is considered.
        """
        try:
            purge = int(purge)
        except ValueError:
            purge = 0
        if purge > 0:
            self.logger.info("Simple purge - remove freestanding snapshots older "\
                             "than %s days." % purge)

            while True:
                _was_removed = False
                snapshots = _get_snapshots_older_than(self.get_snapshots(), purge)
                for snp in snapshots:
                    if snp.getName() == no_purge_snp:
                        self.logger.debug("`%s` skipped.")
                        continue

                    self.logger.debug("Checking '%s' for childs." % (snp))
                    childs = self._retrieve_childsnps(snapshot = snp)
                    if len(childs) == 0:
                        self.logger.debug("Snapshot '%s' has no childs "\
                                        "-> is being removed." % (snp))
                        self.remove_snapshot_forced(snapshot = snp)
                        _was_removed = True
                        break
                if _was_removed is not True:
                    break


def _get_snapshots_younger_than(snapshots, age):
    _res = []
    for snp in snapshots:
        date = snp.getDate()
        snp_age = (datetime.date.today() - datetime.date(date['year'],
                                                         date['month'],
                                                         date['day'])).days
        if snp_age < age:
            _res.append(snp)
    return _res


def _get_snapshots_older_than(snapshots, age):
    _res = []
    for snp in snapshots:
        date = snp.getDate()
        snp_age = (datetime.date.today() - datetime.date(date['year'],
                                                    date['month'],
                                                    date['day'])).days
        if snp_age > age:
            _res.append(snp)
    return _res


def debug_print_snarfile(filename):
    """Print function only for debugging.
    
    :param filename: full path of snar to be printed out
    :type filename: string
    
    """
    _fop = fam.get_file_operations_facade_instance()
    if _fop.path_exists(filename):
        _snar = SnapshotFile(filename, writeFlag = False)
        print "\nSUMMARY of SNAR '%s':" % filename
        for _record in _snar.parseFormat2():
            print "%s" % _record
    else:
        print "\nSUMMARY of SNAR '%s': file not found!" % filename


def debug_snarfile_to_list(filename):
    """Helper function for debugging: the snar-file given by parameter
    'filename' is converted into a list and this list is returned by
    the function.

    :param filename: full path of snar to be converted
    :type filename: string
    :return: list containing snar file entries
    
    """
    _res = []
    _fop = fam.get_file_operations_facade_instance()
    if _fop.path_exists(filename):
        _snar = SnapshotFile(filename, writeFlag = False)
        for _record in _snar.parseFormat2():
            _res.append(_record)
    return _res



#   Simple Backup - Indicator application (status icon)
#                   targeting Ubuntu >= 10.04
#
#   Copyright (c)2009-2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import sys
import appindicator #@UnresolvedImport


from sbackup import util
from sbackup.util.log import LogFactory
from sbackup.util import constants

from sbackup.ui import indicator_core
from sbackup.ui.indicator_core import PyNotifyMixin, SBackupdIndicatorHandler, SBackupdIndicatorBase
from sbackup.ui import misc
from sbackup.util import system


sys.excepthook = misc.except_hook
system.launch_dbus_if_required()


class SBackupdIndicator(SBackupdIndicatorBase, PyNotifyMixin):
    """Graphical front-end in status tray.
    """

    def __init__(self, indicator_hdl):
        if not isinstance(indicator_hdl, SBackupdIndicatorHandler):
            raise TypeError("Parameter of type `SBackupdIndicatorHandler` expected.")

        SBackupdIndicatorBase.__init__(self, indicator_hdl)

        self.logger = LogFactory.getLogger()
        self._indicator = appindicator.Indicator(constants.INDICATORAPP_NAME,
                                                 constants.INDICATOR_ACTIVE_ICON,
                                                 appindicator.CATEGORY_APPLICATION_STATUS)

        PyNotifyMixin.__init__(self, logger = self.logger,
                               iconfile = util.get_resource_file(constants.NOTIFICATION_ICON_FILENAME),
                               trayicon = None)

        self.__init_ctrls()

    def _notify_info(self, profilename, message):
        PyNotifyMixin._notify_info(self, profilename, message)

    def _notify_warning(self, profilename, message):
        PyNotifyMixin._notify_warning(self, profilename, message)

    def _notify_error(self, profilename, message):
        PyNotifyMixin._notify_error(self, profilename, message)

    def __init_ctrls(self):
        self._indicator.set_status(appindicator.STATUS_ACTIVE)
        self._indicator.set_attention_icon(constants.INDICATOR_ATTENTION_ICON)

        self._build_menu()
# TODO: Add option to dictionary: show item initially.
        self._indicator.set_menu(self._menu)
        for _item in self._menuitems:
            self._menuitems[_item].show()
        self._show_showdialogs_menuitem()

    def set_status_to_normal(self):
        self._indicator.set_icon(constants.INDICATOR_ACTIVE_ICON)
        self._indicator.set_status(appindicator.STATUS_ACTIVE)

    def set_status_to_attention(self):
        self._indicator.set_status(appindicator.STATUS_ATTENTION)

    def set_status_to_finished(self):
        self._indicator.set_icon(constants.INDICATOR_SUCCESS_ICON)


def main(options):
    exitcode = indicator_core.main(options, SBackupdIndicator)
    return exitcode



#   Simple Backup - Logging facilities
#
#   Copyright (c)2008-2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#   Copyright (c)2007-2009: Ouattara Oumar Aziz <wattazoum@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#


import logging
import os.path


def shutdown_logging():
    logging.shutdown()


class LogFactory(object):
    """
    """
    logger = None
    created_loggers = []

    #create formatter
    formatter = logging.Formatter("%(asctime)s - %(levelname)s: %(message)s")
    debug_formatter = logging.Formatter("%(asctime)s - %(levelname)s in "\
                        "%(module)s.%(funcName)s(%(lineno)d): %(message)s")

    def __init__(self):
        pass

    def getLogger(name = None, logfile = None, level = 20) :
        """Returns last used logger instance. If no instance exists, a new
        one is created.
        
        @param name: The name of the logger
        @param logfile : default=False
        @param level: The level of the logger (default = logging.INFO(20) )
        """
        if LogFactory.logger :
            if name:
                if LogFactory.logger.name == name :
                    return LogFactory.logger
                else :
                    return LogFactory.__createLogger(name, logfile, level)
            else :
                return LogFactory.logger
        else :
            return LogFactory.__createLogger(name, logfile, level)

    getLogger = staticmethod(getLogger)

    def __createLogger(name = None, logfile = None, level = 20):
        """Private helper method that creates a new logger instance.
        To avoid the overwriting of previous settings, the names of
        already created loggers are stored.
        """
        if not name:
            name = "SBackup"

        _formatter = LogFactory.formatter
        if level == logging.DEBUG:
            _formatter = LogFactory.debug_formatter

        #create logger
        LogFactory.logger = logging.getLogger(name)
        if name in LogFactory.created_loggers:
            pass
        else:
            LogFactory.created_loggers.append(name)
            LogFactory.logger.setLevel(level)

            #create console handler and set level and formatter
            ch = logging.StreamHandler()
            ch.setLevel(level)
            ch.setFormatter(_formatter)
            LogFactory.logger.addHandler(ch)

            if logfile:
                # create the logfile
                if not os.path.exists(logfile) :
                    #make sure that the parent directory exist
                    parentdir = os.path.dirname(os.path.abspath(logfile))
                    if not os.path.exists(parentdir):
                        os.makedirs(parentdir)
                    _writetofile(logfile, "SBackup '%s' Logger\r\n==============\r\n" % name)
                else :
                    # clean the logfile
                    os.rename(logfile, logfile + ".old")
                    _writetofile(logfile, "SBackup '%s' Logger\r\n==============\r\n" % name)
                ch1 = logging.FileHandler(logfile)
                ch1.setLevel(level)
                ch1.setFormatter(_formatter)
                LogFactory.logger.addHandler(ch1)
        return LogFactory.logger

    __createLogger = staticmethod(__createLogger)


def _writetofile(path, content) :
    _fobj = open(path, "w")
    _fobj.write(content)
    _fobj.close()



#   Simple Backup - handling of restoration processes
#
#   Copyright (c)2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#   Copyright (c)2007-2008: Ouattara Oumar Aziz <wattazoum@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from gettext import gettext as _
import tempfile
import datetime


from sbackup.core import SnapshotManager
from sbackup.fs_backend import fam

from sbackup.util import exceptions
from sbackup.util import log
from sbackup.ar_backend import tar


class RestoreManager(object):
    """
    """

    def __init__(self, targethandler):
        """
        """
        self.logger = log.LogFactory.getLogger()
        self.__fam_target_hdl = targethandler
        self._fop = fam.get_file_operations_facade_instance()

    def restore(self, snapshot, _file):
        """
        Restore one file or directory from the backup tdir with name
        file to its old location.
        All existing files must be moved to a "*.before_restore_$time" files.
        @param snapshot: 
        @param file : 
        """
        self.restoreAs(snapshot, _file, None)

    def restoreAs(self, snapshot, _file, target, backupFlag = True, failOnNotFound = True) :
        """
        Restore one file or directory from the backup tdir with name
        file to target (or to its old location if None if given to target).
        All existing files must be moved to a "*.before_restore_$time" files.
        @param snapshot:
        @param file :  
        @param target: where to restore given file/dir (not the backup destination/snapshot dir)
        @param backupFlag: Set to false to make no backup when restoring (default = True)
        @param failOnNotFound: set to False if we don't want to fail if a file is not found (default is True)
        
        :todo: Re-factor and simplify!
        """
        if snapshot is None:
            raise exceptions.SBException("Please provide a Snapshot")
        if _file is None:
            raise exceptions.SBException("Please provide a File/directory")

        _file = "%s%s" % (self._fop.pathsep, _file.lstrip(self._fop.pathsep))

        self.logger.debug("Restore as\n\tsnapshot: `%s`\n\tfile (path in snapshot): `%s`\n\trestore target: `%s`"\
                          % (snapshot, _file, target))

        if not snapshot.getSnapshotFileInfos().hasPath(_file) and not snapshot.getSnapshotFileInfos().hasFile(_file):
            if failOnNotFound:
                raise exceptions.SBException(_("File '%s' not found in the backup snapshot files list") % _file)
            else:
                self.logger.warning(_("File '%(filename)s' not found in snapshot's [%(snapshotname)s] files list, Skipped.")\
                                    % {"filename": _file, "snapshotname" : snapshot.getName()})
                return

        _snpname = self._fop.get_basename(snapshot.getName())
        _arname = self._fop.get_basename(snapshot.getArchive())
        _larpath = self.__fam_target_hdl.get_eff_fullpath(_snpname, _arname)
        self.logger.debug("eff. local path to archive: %s" % _larpath)

        suffix = None
        if backupFlag :
            now = datetime.datetime.now().isoformat("_").replace(":", ".")
            suffix = ".before_restore_" + now

        if target and self._fop.path_exists(target):
            self.logger.debug("Restore target is given and exists")
            if self._fop.is_dir(target) is True:
                self.logger.debug("Restore target is a directory (i.e. we extract into a directory)")

                #create a temp dir, extract inside then move the content
                _tmpdir = tempfile.mkdtemp(dir = target, prefix = 'sbackup-restore_')
                self.logger.debug("Restore tempdir: `%s`" % _tmpdir)

                tar.extract(snapshot.getArchive(), _larpath, _file, _tmpdir,
                            bckupsuffix = suffix, splitsize = snapshot.getSplitedSize())

                _file_in_target = self._fop.joinpath(target, self._fop.get_basename(_file))
                _file_in_tmpdir = self._fop.joinpath(_tmpdir, _file)

                self.logger.debug("File in restore target: `%s`" % _file_in_target)
                self.logger.debug("File in restore tempdir: `%s`" % _file_in_tmpdir)

                if backupFlag:
                    if self._fop.path_exists(_file_in_target):
                        self._fop.force_move(_file_in_target, "%s%s" % (_file_in_target, suffix))

                self._fop.force_move(_file_in_tmpdir, _file_in_target)
                self._fop.force_delete(_tmpdir)
            else:
                #the target is a file (i.e. the backuped file is restored under new name
                parent = self._fop.get_dirname(target)
                tar.extract(snapshot.getArchive(), _larpath, _file, parent,
                            bckupsuffix = suffix, splitsize = snapshot.getSplitedSize())
        else:
            # target is set to None or target not exists
            if target and not self._fop.path_exists(target) :
                #target != None but target doesn't exists
                self._fop.makedirs(target)
                tar.extract(snapshot.getArchive(), _larpath, _file, target,
                            splitsize = snapshot.getSplitedSize())
            else :
                # Target = None , extract at the place it belongs
                if self._fop.path_exists(_file) :
                    # file exist:
                    tar.extract(snapshot.getArchive(), _larpath, _file, target,
                                bckupsuffix = suffix, splitsize = snapshot.getSplitedSize())
                else :
                    # file doesn't exist nothing to move, just extract
                    tar.extract(snapshot.getArchive(), _larpath, _file, target,
                                splitsize = snapshot.getSplitedSize())

    def revert(self, snapshot, directory):
        """
        Revert a directory to its snapshot date state.
        @param snapshot : The snapshot from which to revert 
        @param dir : the dir to revert, use self._fop.pathsep for the whole snapshot
        """
        self.revertAs(snapshot, directory, None)


    def __cleanBackupedFiles(self, directory , suffix):
        """
        clean the backuped copies in the directory (dir) that ends with suffix
        @param dir: directory to clean up
        @param suffix: the suffix of backuped files
        """
        pass


    def revertAs(self, snapshot, directory, targetdir):
        """
        Revert a directory to its snapshot date state into a directory.
        We will restore the directory starting from the base snapshot to the selected one and clean the restore directory each time.
        @param snapshot : The snapshot from which to revert 
        @param dir : the dir to revert, use self._fop.pathsep for the whole snapshot
        @param targetdir: The dir in which to restore files 
        """
        if not snapshot:
            raise exceptions.SBException("Please provide a Snapshot")
        if not directory:
            raise exceptions.SBException("Please provide a File/directory")

        snpman = SnapshotManager.SnapshotManager(self._fop.get_dirname(snapshot.getPath()))
        history = snpman.getSnpHistory(snapshot)
        history.reverse()
        _cnt = 0
        for snp in history:
            self.logger.debug("Restoring '%(dirname)s' from snapshot '%(snapshotname)s'"\
                              % { "dirname" : directory, "snapshotname" : snp.getName() })
            _bak = False
            if _cnt == 0:
                _bak = True
            _cnt += 1

            self.restoreAs(snp, directory, targetdir, backupFlag = _bak,
                           failOnNotFound = False)

